"""File vns.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 16 16:55:30 2020

Multi-objective variable neighborhood search algorithm
"""

# pylint: disable=invalid-name, too-many-statements, too-many-arguments, too-many-branches
# pylint: disable=too-many-locals
import random
import abc
import pickle
import numpy as np
import biogeme.exceptions as excep
import biogeme.messaging as msg

logger = msg.bioMessage()


class problemClass(metaclass=abc.ABCMeta):
    """
    Abstract class defining a problem
    """

    @abc.abstractmethod
    def isValid(self, aSolution):
        """Evaluate  the validity of the solution.

        :param aSolution: solution to be checked
        :type aSolution: :class:`biogeme.vns.solutionClass`

        :return: valid, why where valid is True if the solution is
            valid, and False otherwise. why contains an explanation why it
            is invalid.
        :rtype: tuple(bool, str)
        """

    @abc.abstractmethod
    def evaluate(self, aSolution):
        """Evaluate the objectives functions of the solution and store them in
        the solution object.

        :param aSolution: solution to be evaluated
        :type aSolution: :class:`biogeme.vns.solutionClass`

        """

    @abc.abstractmethod
    def describe(self, aSolution):
        """Short description of the solution. Used for reporting.

        :param aSolution: solution to be described
        :type aSolution: :class:`biogeme.vns.solutionClass`

        :return: short description of the solution.
        :rtype: str
        """

    @abc.abstractmethod
    def generateNeighbor(self, aSolution, neighborhoodSize):
        """
        Generates a neighbor of the solution.

        :param aSolution: solution to be modified
        :type aSolution: :class:`biogeme.vns.solutionClass`

        :param neighborhoodSize: size of the neighborhood to be applied
        :type neighborhoodSize: int

        :return: a neighbor solution, and the number of changes that have been
                 actually applied.
        :rtype: tuple(:class:`biogeme.vns.solutionClass`, int)
        """

    @abc.abstractmethod
    def neighborRejected(self, aSolution, aNeighbor):
        """Notify that a neighbor has been rejected by the
        algorithm. Used to update the statistics on the operators.

        :param aSolution: solution  modified
        :type aSolution: :class:`biogeme.vns.solutionClass`

        :param aNeighbor: neighbor
        :type aNeighbor: :class:`biogeme.vns.solutionClass`
        """

    @abc.abstractmethod
    def neighborAccepted(self, aSolution, aNeighbor):
        """Notify that a neighbor has been accepted by the
        algorithm. Used to update the statistics on the operators.

        :param aSolution: solution modified
        :type aSolution: :class:`biogeme.vns.solutionClass`

        :param aNeighbor: neighbor
        :type aNeighbor: :class:`biogeme.vns.solutionClass`
        """


class solutionClass(metaclass=abc.ABCMeta):
    """
    This is an abstract class defining a solution.
    """

    def __init__(self):
        self.objectives = None
        self.objectivesNames = None

    @abc.abstractmethod
    def __repr__(self):
        pass

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    @abc.abstractmethod
    def __str__(self):
        pass

    def dominates(self, anotherSolution):
        """Checks if the current solution dominates another one. It is
        assumed that all objectives are minimized.

        :param anotherSolution: the other solution to be compared with.
        :type anotherSolution: solutionClass

        :return: True is self dominates anotherSolution, False otherwise.
        :rtype: bool

        :raise biogemeError: if the value of no value is availaboe for
            the objective functions of ego.

        :raise biogemeError: if the value of no value is availaboe for
            the objective functions of alter.

        :raise biogemeError: if the number of objective functions are
            incompatible.

        """
        if self.objectives[0] is None:
            raise excep.biogemeError(
                'No values are available for '
                'the objective functions of ego.'
            )

        if anotherSolution.objectives[0] is None:
            raise excep.biogemeError(
                'No values are available for '
                'the objective functions of alter.'
            )

        if len(self.objectives) != len(anotherSolution.objectives):
            raise excep.biogemeError(
                f'Incompatible sizes: '
                f'{len(self.objectives)}'
                f' and {len(anotherSolution.objectives)}'
            )

        for my_f, her_f in zip(self.objectives, anotherSolution.objectives):
            if my_f > her_f:
                return False

        return self.objectives != anotherSolution.objectives


class operatorsManagement:
    """
    Class managing the selection and performance analysis of the operators
    """

    def __init__(self, operatorNames):
        """
        :param operatorNames: names of operators
        :type operators: list(str)
        """
        self.scores = {k: 0 for k in operatorNames}
        """ dict of scores obtained by the operators"""

        self.names = list(self.scores)
        """ Names of the operators """

        self.available = {k: True for k in operatorNames}
        """ dict of availability of the operators """

    def increaseScore(self, operator):
        """Increase the score of an operator.

        :param operator: name of the operator
        :type operator: str

        :raise biogemeError: if the operator is not known.
        """
        try:
            self.scores[operator] += 1
        except KeyError as e:
            raise excep.biogemeError(f'Unknown operator: {operator}') from e

    def decreaseScore(self, operator):
        """Decrease the score of an operator. If it has already the minimum
        score, it increases the others.

        :param operator: name of the operator
        :type operator: str

        :raise biogemeError: if the operator is not known.
        """
        try:
            self.scores[operator] -= 1
        except KeyError as e:
            raise excep.biogemeError(f'Unknown operator: {operator}') from e

    def selectOperator(self, minProbability=0.01, scale=0.1):
        """Select an operator based on the scores

        :param minProbability: minimum probability to select any
                               operator. This is meant to avoid
                               degeneracy, that is to have operators
                               that have no chance to be chosen

        :type minProbability: float

        :param scale: parameter for the choice probability
        :type scale: float

        :return: name of the selected operator
        :rtype: str

        :raise biogemeError: if the minimum probability is too large
            for the number of operators. It must be less or equal to 1 /
            N, where N is the number of operators.

        """

        if minProbability > 1.0 / len(self.scores):
            raise excep.biogemeError(
                f'Cannot impose min. probability '
                f'{minProbability} '
                f'with {len(self.scores)} operators'
            )

        maxscore = max(list(self.scores.values()))
        listOfScores = np.array(
            [
                np.exp(scale * (s - maxscore)) if self.available[k] else 0
                for k, s in self.scores.items()
            ]
        )
        if len(listOfScores) == 0:
            return None
        totalScore = sum(listOfScores)
        prob = np.array([float(s) / float(totalScore) for s in listOfScores])
        # Enforce minimum probability
        ok = prob >= minProbability
        tooLow = prob < minProbability
        notzero = prob != 0.0
        update = tooLow & notzero

        reservedTotal = update.sum() * minProbability
        totalHighScores = listOfScores[ok].sum()
        prob[ok] = (1.0 - reservedTotal) * listOfScores[ok] / totalHighScores
        prob[update] = minProbability
        return np.random.choice(self.names, 1, p=prob)[0]


class paretoClass:
    """Class managing the solutions"""

    def __init__(self, maxNeighborhood, archiveInputFile=None):
        """

        :param maxNeighborhood: the maximum size of the neighborhood
            that must be considered
        :type maxNeighborhood: int

        :param archiveInputFile: name of a pickle file contaning sets
                                 from previous runs
        :type archiveInputFile: str

        """

        self.maxNeighborhood = maxNeighborhood
        """the maximum size of the neighborhood that must be considered
        """

        if archiveInputFile is None:
            self.pareto = dict()
            """dict containing the pareto set. The keys are the solutions
            (class :class:`biogeme.vns.solutionClass`), and the values are
            the size of the neighborhood that must be applied the next
            time tat the solution is selected by the algorithm.

            """

            self.removed = set()
            """set of solutions that have been in the Pareto set ar some point,
            but have been removed because dominated by another
            solution.
            """
            self.considered = set()
            """set of solutions that have been considered at some point by the
            algorithm
            """
        else:
            if self.restore(archiveInputFile):
                logger.general(
                    f'Pareto set initialized from file with '
                    f'{self.length()} elements.'
                )
            else:
                logger.general(
                    f'Unable to read file {archiveInputFile}. '
                    f'Pareto set empty.'
                )
                self.pareto = dict()
                self.removed = set()
                self.considered = set()

    def length(self):
        """
        Obtain the length of the parato set.
        """
        return len(self.pareto)

    def changeNeighborhood(self, sol):
        """
        Change the size of the neighborhood for a solution in the Pareto set.

        :param sol: solution for which the neighborhood size must be increased.
        :type sol: solutionClass
        """
        if sol not in self.pareto:
            return
        self.pareto[sol] += 1

    def resetNeighborhood(self, sol):
        """
        Reset the size of the neighborhood to 1 for a solution.

        :param sol: solution for which the neighborhood size must be reset.
        :type sol: solutionClass
        """
        if sol not in self.pareto:
            return
        self.pareto[sol] = 1

    def add(self, elem):
        """

        - We define the set D as the set of members of the current
          Pareto set that are dominated by the element elem.

        - We define the set S as the set of members of the current
          Pareto set that dominate the element elem.

        If S is empty, we add elem to the Pareto set, and remove all
        elements in D.

        :param elem: elem to be considered for inclusion in the Pareto set.
        :type elem: solutionClass

        :return: True is elem has been included. False otherwise.
        :rtype: bool

        """
        self.considered.add(elem)
        S_dominated = set()
        D_dominating = set()
        for k in self.pareto:
            if elem.dominates(k):
                D_dominating.add(k)
            if k.dominates(elem):
                S_dominated.add(k)
        if S_dominated:
            return False
        self.pareto[elem] = 1
        self.pareto = {k: v for k, v in self.pareto.items()
                       if k not in D_dominating}
        self.removed |= D_dominating
        return True

    def select(self):
        """
        Select a candidate to be modified during the next iteration.

        :return: a candidate and the neghborhoodsize
        :rtype: tuple(solutionClass, int)

        """

        # Candidates are members of the Pareto set that have not
        # reached the maximum neighborhood size.

        candidates = [
            (k, v) for k, v in self.pareto.items() if v < self.maxNeighborhood
        ]

        if not candidates:
            return None, None

        aSolution, neighborhoodSize = random.choice(candidates)

        return aSolution, neighborhoodSize

    def dump(self, fileName):
        """
        Dump the three sets on a file using pickle

        :param fileName: name of the file where the sets must be saved.
        :type fileName: str

        :raise biogemeError: if a problem has occured during dumping.
        """

        allSets = self.pareto, self.removed, self.considered
        with open(fileName, 'wb') as f:
            pickle.dump(allSets, f)

        with open(fileName, 'rb') as f:
            tmppareto, _, _ = pickle.load(f)
        if len(tmppareto) != len(self.pareto):
            raise excep.biogemeError(
                f'Problem in dumping: {len(tmppareto)} '
                f'instead of {len(self.pareto)}'
            )

    def restore(self, fileName):
        """Restore the Pareto from a file"""
        try:
            with open(fileName, 'rb') as f:
                self.pareto, self.removed, self.considered = pickle.load(f)
        except FileNotFoundError:
            return False

        return True


def vns(
    problem,
    firstSolutions,
    maxNeighborhood=20,
    numberOfNeighbors=10,
    archiveInputFile=None,
    pickleOutputFile=None,
):
    """Multi objective Variable Neighborhood Search

    :param problem: problem description
    :type problem: problemClass

    :param firstSolutions: several models to initialize  the algorithm
    :type firstSolutions: list(solutionClass)

    :param maxNeighborhood: the maximum size of the neighborhood
        that must be considered. Default: 20
    :type maxNeighborhood: int

    :param numberOfNeighbors: if none of this neighbors improves the
                              solution, it is considered that a local
                              optimum has been reached.
    :type numberOfNeighbors: int

    :param archiveInputFile: name of pickle file containing
                             specifications from a previous run of the
                             algorithm. Default: None
    :type archiveInputFile: str

    :param pickleOutputFile: name of file where the set of models are stored:
    :type pickleOutputFile: str

    :return: the pareto set, the set of models that have been in the
             pareto set and then removed, and the set of all models
             considered by the algorithm.
    :rtype: class :class:`biogeme.vns.paretoClass`

    :raise biogemeError: if the first Pareto set is empty.

    """

    if pickleOutputFile is None:
        pickleOutputFile = '__defaultPareto.pickle'

    thePareto = paretoClass(maxNeighborhood, archiveInputFile)

    if firstSolutions is not None:
        for s in firstSolutions:
            valid, why = problem.isValid(s)
            if valid:
                problem.evaluate(s)
                thePareto.add(s)
                logger.detailed(problem.describe(s))
            else:
                logger.warning(problem.describe(s))
                logger.warning(f'Default specification is invalid: {why}')

    if thePareto.length() == 0:
        raise excep.biogemeError(
            'Cannot start the algorithm ' 'with an empty Pareto set.'
        )

    logger.general(f'Initial pareto: {thePareto.length()}')
    thePareto.dump(pickleOutputFile)

    solutionToImprove, neighborhoodSize = thePareto.select()

    while solutionToImprove is not None:
        continueWithSolution = True
        n = 0
        while continueWithSolution:
            logger.general(f'----> Neighbor {n} of size {neighborhoodSize}')
            aNeighbor, numberOfChanges = problem.generateNeighbor(
                solutionToImprove, neighborhoodSize
            )

            n += 1
            if n == numberOfNeighbors:
                thePareto.changeNeighborhood(solutionToImprove)
                continueWithSolution = False

            if numberOfChanges == 0:
                logger.warning(
                    f'Neighbor of size {neighborhoodSize}: '
                    f'generation failed'
                )
                thePareto.changeNeighborhood(solutionToImprove)
                continueWithSolution = False
            elif aNeighbor in thePareto.considered:
                problem.neighborRejected(solutionToImprove, aNeighbor)
                logger.general(
                    f'*** Neighbor of size {neighborhoodSize}:'
                    f' already considered***'
                )
                thePareto.changeNeighborhood(solutionToImprove)
            else:
                problem.evaluate(aNeighbor)
                valid, why = problem.isValid(aNeighbor)
                if valid:
                    if thePareto.add(aNeighbor):
                        logger.general('*** New pareto solution: ')
                        logger.detailed(problem.describe(aNeighbor))
                        problem.neighborAccepted(solutionToImprove, aNeighbor)
                        thePareto.dump(pickleOutputFile)
                        thePareto.resetNeighborhood(solutionToImprove)
                        continueWithSolution = False
                    else:
                        logger.general(
                            f'*** Neighbor of size '
                            f'{neighborhoodSize}: dominated***'
                        )
                        problem.neighborRejected(solutionToImprove, aNeighbor)
                        thePareto.changeNeighborhood(solutionToImprove)
                else:
                    logger.general(
                        f'*** Neighbor of size {neighborhoodSize}'
                        f' invalid: {why}***'
                    )
                    problem.neighborRejected(solutionToImprove, aNeighbor)
                    thePareto.changeNeighborhood(solutionToImprove)

        solutionToImprove, neighborhoodSize = thePareto.select()

    return thePareto



"""
Test the draws module

:author: Michel Bierlaire
:data: Wed Apr 29 17:56:55 2020
"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import unittest
import biogeme.draws as dr
import numpy as np

class testDraws(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)

    def test_getUniform(self):
        draws = dr.getUniform(sampleSize=5, numberOfDraws=10, symmetric=False)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > 0)
        self.assertTrue(np.max(draws) <= 1)

    def test_getUniformSymmetric(self):
        draws = dr.getUniform(sampleSize=5, numberOfDraws=10, symmetric=True)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > -1)
        self.assertTrue(np.max(draws) <= 1)

    def test_getLatinHypercubeDraws(self):
        draws = dr.getLatinHypercubeDraws(sampleSize=5, numberOfDraws=10)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > 0)
        self.assertTrue(np.max(draws) <= 1)

    def test_getLatinHypercubeDrawsSymmetric(self):
        draws = dr.getLatinHypercubeDraws(sampleSize=5, numberOfDraws=10, symmetric=True)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > -1)
        self.assertTrue(np.max(draws) <= 1)

    def test_userDefinedDraws(self):
        myUnif = np.random.uniform(size=30)
        draws = dr.getLatinHypercubeDraws(sampleSize=3,
                                          numberOfDraws=10,
                                          symmetric=False,
                                          uniformNumbers=myUnif)
        r, c = draws.shape
        self.assertEqual(r, 3)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > 0)
        self.assertTrue(np.max(draws) <= 1)

    def test_halton(self):
        halton = dr.getHaltonDraws(sampleSize=1, numberOfDraws=100, base=3)
        h = np.array([0.33333333,
                      0.66666667,
                      0.11111111,
                      0.44444444,
                      0.77777778,
                      0.22222222,
                      0.55555556,
                      0.88888889,
                      0.03703704,
                      0.37037037,
                      0.7037037,
                      0.14814815,
                      0.48148148,
                      0.81481481,
                      0.25925926,
                      0.59259259,
                      0.92592593,
                      0.07407407,
                      0.40740741,
                      0.74074074,
                      0.18518519,
                      0.51851852,
                      0.85185185,
                      0.2962963,
                      0.62962963,
                      0.96296296,
                      0.01234568,
                      0.34567901,
                      0.67901235,
                      0.12345679,
                      0.45679012,
                      0.79012346,
                      0.2345679,
                      0.56790123,
                      0.90123457,
                      0.04938272,
                      0.38271605,
                      0.71604938,
                      0.16049383,
                      0.49382716,
                      0.82716049,
                      0.27160494,
                      0.60493827,
                      0.9382716,
                      0.08641975,
                      0.41975309,
                      0.75308642,
                      0.19753086,
                      0.5308642,
                      0.86419753,
                      0.30864198,
                      0.64197531,
                      0.97530864,
                      0.02469136,
                      0.35802469,
                      0.69135802,
                      0.13580247,
                      0.4691358,
                      0.80246914,
                      0.24691358,
                      0.58024691,
                      0.91358025,
                      0.0617284,
                      0.39506173,
                      0.72839506,
                      0.17283951,
                      0.50617284,
                      0.83950617,
                      0.28395062,
                      0.61728395,
                      0.95061728,
                      0.09876543,
                      0.43209877,
                      0.7654321,
                      0.20987654,
                      0.54320988,
                      0.87654321,
                      0.32098765,
                      0.65432099,
                      0.98765432,
                      0.00411523,
                      0.33744856,
                      0.67078189,
                      0.11522634,
                      0.44855967,
                      0.781893,
                      0.22633745,
                      0.55967078,
                      0.89300412,
                      0.04115226,
                      0.3744856,
                      0.70781893,
                      0.15226337,
                      0.48559671,
                      0.81893004,
                      0.26337449,
                      0.59670782,
                      0.93004115,
                      0.0781893,
                      0.41152263])
        diff = halton - h
        norm = np.linalg.norm(diff)
        self.assertAlmostEqual(norm, 0, 7)
        halton = dr.getHaltonDraws(sampleSize=1, numberOfDraws=10, base=3, skip=10)
        h = [0.7037037,
             0.14814815,
             0.48148148,
             0.81481481,
             0.25925926,
             0.59259259,
             0.92592593,
             0.07407407,
             0.40740741,
             0.74074074]
        diff = halton - h
        norm = np.linalg.norm(diff)
        self.assertAlmostEqual(norm, 0, 7)

    def test_antithetic(self):
        draws = dr.getAntithetic(dr.getHaltonDraws, sampleSize=1, numberOfDraws=10)
        d = [0.5,
             0.25,
             0.75,
             0.125,
             0.625,
             0.5,
             0.75,
             0.25,
             0.875,
             0.375]
        self.assertListEqual(d, draws[0].tolist())

    def test_normal(self):
        draws = dr.getNormalWichuraDraws(sampleSize=3, numberOfDraws=1000000)
        mean = np.linalg.norm(np.average(draws, axis=1))
        self.assertAlmostEqual(mean, 0, 2)
        draws = dr.getNormalWichuraDraws(sampleSize=3, numberOfDraws=1000000, antithetic=True)
        mean = np.linalg.norm(np.average(draws, axis=1))
        self.assertAlmostEqual(mean, 0, 5)

if __name__ == '__main__':
    unittest.main()



"""File m03_simultaneous_estimation.py

Choice model with the latent variable.
Mixture of logit.
Measurement equation for the indicators.
Maximum likelihood (full information) estimation.

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 08:22:28 2019
"""

import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
import biogeme.results as res
import biogeme.optimization as opt
import biogeme.messaging as msg
from biogeme import models
from biogeme.expressions import (
    Beta,
    DefineVariable,
    Variable,
    log,
    RandomVariable,
    Integrate,
    Elem,
    bioNormalCdf,
    exp,
)

# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# The following statement allows you to use the names of the variable
# as Python variable.
Choice = Variable('Choice')
Gender = Variable('Gender')
FamilSitu = Variable('FamilSitu')
Education = Variable('Education')
ResidChild = Variable('ResidChild')
SocioProfCat = Variable('SocioProfCat')
age = Variable('age')
TimePT = Variable('TimePT')
TimeCar = Variable('TimeCar')
MarginalCostPT = Variable('MarginalCostPT')
CostCarCHF = Variable('CostCarCHF')
distance_km = Variable('distance_km')
TripPurpose = Variable('TripPurpose')
WaitingTimePT = Variable('WaitingTimePT')
Choice = Variable('Choice')


# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Read the estimates from the structural equation estimation
NAME = 'm01_latent_variable'
try:
    structResults = res.bioResults(pickleFile=f'{NAME}.pickle')
except FileNotFoundError:
    print(
        f'Run first the script {NAME}.py in order to generate the file '
        f'{NAME}.pickle.'
    )
    sys.exit()
structBetas = structResults.getBetaValues()


### Variables


# Definition of other variables
male = DefineVariable('male', Gender == 1, database)
haveChildren = DefineVariable(
    'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0, database
)
highEducation = DefineVariable('highEducation', Education >= 6, database)
childCenter = DefineVariable(
    'childCenter', ((ResidChild == 1) + (ResidChild == 2)) > 0, database
)
childSuburb = DefineVariable(
    'childSuburb', ((ResidChild == 3) + (ResidChild == 4)) > 0, database
)

### Coefficients

coef_intercept = Beta(
    'coef_intercept', structBetas['coef_intercept'], None, None, 0
)
coef_age_30_less = Beta(
    'coef_age_30_less', structBetas['coef_age_30_less'], None, None, 0
)
coef_haveChildren = Beta(
    'coef_haveChildren', structBetas['coef_haveChildren'], None, None, 0
)

coef_highEducation = Beta(
    'coef_highEducation', structBetas['coef_highEducation'], None, None, 0
)

coef_artisans = Beta(
    'coef_artisans', structBetas['coef_artisans'], None, None, 0
)

coef_employees = Beta(
    'coef_employees', structBetas['coef_employees'], None, None, 0
)

coef_male = Beta('coef_male', structBetas['coef_male'], None, None, 0)

coef_child_center = Beta(
    'coef_child_center', structBetas['coef_child_center'], None, None, 0
)
coef_child_suburb = Beta(
    'coef_child_suburb', structBetas['coef_child_suburb'], None, None, 0
)

### Latent variable: structural equation

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

ACTIVELIFE = (
    coef_intercept
    + coef_child_center * childCenter
    + coef_child_suburb * childSuburb
    + coef_highEducation * highEducation
    + coef_artisans * (SocioProfCat == 5)
    + coef_employees * (SocioProfCat == 6)
    + coef_age_30_less * (age <= 30)
    + coef_male * male
    + coef_haveChildren * haveChildren
    + sigma_s * omega
)

### Measurement equations

indicators = [
    'ResidCh01',
    'ResidCh04',
    'ResidCh05',
    'ResidCh06',
    'LifSty07',
    'LifSty10',
]

# We define the intercept parameters. The first one is normalized to 0.
INTER = {
    k: Beta(f'INTER_{k}', structBetas[f'INTER_{k}'], None, None, 0)
    for k in indicators[1:]
}
INTER[indicators[0]] = Beta(f'INTER_{indicators[0]}', 0, None, None, 1)

# We define the coefficients. The first one is normalized to 1.
B = {
    k: Beta(f'B_{k}', structBetas[f'B_{k}'], None, None, 0)
    for k in indicators[1:]
}
B[indicators[0]] = Beta(f'B_{indicators[0]}', 1, None, None, 1)

# We define the measurement equations for each indicator
MODEL = {k: INTER[k] + B[k] * ACTIVELIFE for k in indicators}

# We define the scale parameters of the error terms.
SIGMA_STAR = {
    k: Beta(f'SIGMA_STAR_{k}', structBetas[f'SIGMA_STAR_{k}'], 1.0e-5, None, 0)
    for k in indicators[1:]
}
SIGMA_STAR[indicators[0]] = Beta(
    f'SIGMA_STAR_{indicators[0]}', 1, None, None, 1
)

delta_1 = Beta('delta_1', structBetas['delta_1'], 1.0e-5, None, 0)
delta_2 = Beta('delta_2', structBetas['delta_2'], 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

tau_1_residual = {k: (tau_1 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
tau_2_residual = {k: (tau_2 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
tau_3_residual = {k: (tau_3 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
tau_4_residual = {k: (tau_4 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
Ind = {
    k: {
        1: bioNormalCdf(tau_1_residual[k]),
        2: bioNormalCdf(tau_2_residual[k]) - bioNormalCdf(tau_1_residual[k]),
        3: bioNormalCdf(tau_3_residual[k]) - bioNormalCdf(tau_2_residual[k]),
        4: bioNormalCdf(tau_4_residual[k]) - bioNormalCdf(tau_3_residual[k]),
        5: 1 - bioNormalCdf(tau_4_residual[k]),
        6: 1.0,
        -1: 1.0,
        -2: 1.0,
    }
    for k in indicators
}

prob_indicators = Elem(Ind[indicators[0]], Variable(indicators[0]))
for k in indicators[1:]:
    prob_indicators *= Elem(Ind[k], Variable(k))

# Choice model
# Read the estimates from the sequential estimation, and use
# them as starting values

NAME_SEQ = 'm02_sequential_estimation'
try:
    choiceResults = res.bioResults(pickleFile=f'{NAME_SEQ}.pickle')
except FileNotFoundError:
    print(
        f'Run first the script {NAME_SEQ}.py in order to generate the file '
        f'{NAME_SEQ}.pickle.'
    )
    sys.exit()

choiceBetas = choiceResults.getBetaValues()

ASC_CAR = Beta('ASC_CAR', choiceBetas['ASC_CAR'], None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', choiceBetas['ASC_SM'], None, None, 0)
BETA_COST_HWH = Beta(
    'BETA_COST_HWH', choiceBetas['BETA_COST_HWH'], None, None, 0
)
BETA_COST_OTHER = Beta(
    'BETA_COST_OTHER', choiceBetas['BETA_COST_OTHER'], None, None, 0
)
BETA_DIST = Beta('BETA_DIST', choiceBetas['BETA_DIST'], None, None, 0)
BETA_TIME_CAR_REF = Beta(
    'BETA_TIME_CAR_REF', choiceBetas['BETA_TIME_CAR_REF'], None, 0, 0
)
BETA_TIME_CAR_AL = Beta(
    'BETA_TIME_CAR_AL', choiceBetas['BETA_TIME_CAR_AL'], None, None, 0
)
BETA_TIME_PT_REF = Beta(
    'BETA_TIME_PT_REF', choiceBetas['BETA_TIME_PT_REF'], None, 0, 0
)
BETA_TIME_PT_AL = Beta(
    'BETA_TIME_PT_AL', choiceBetas['BETA_TIME_PT_AL'], None, None, 0
)
BETA_WAITING_TIME = Beta(
    'BETA_WAITING_TIME', choiceBetas['BETA_WAITING_TIME'], None, None, 0
)

TimePT_scaled = DefineVariable('TimePT_scaled', TimePT / 200, database)
TimeCar_scaled = DefineVariable('TimeCar_scaled', TimeCar / 200, database)
MarginalCostPT_scaled = DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10, database
)
CostCarCHF_scaled = DefineVariable(
    'CostCarCHF_scaled', CostCarCHF / 10, database
)
distance_km_scaled = DefineVariable(
    'distance_km_scaled', distance_km / 5, database
)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1, database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1, database)

### Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_AL * ACTIVELIFE)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_AL * ACTIVELIFE)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel) for
# the choice
condprob = models.logit(V, None, Choice)

# Conditional to omega, we have the product of ordered probit for the
# indicators.
condlike = prob_indicators * condprob

# We integrate over omega using numerical integration
loglike = log(Integrate(condlike * density, 'omega'))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, loglike)
biogeme.modelName = 'm03_simultaneous_estimation'

# Estimate the parameters
results = biogeme.estimate(algorithm=opt.bioBfgs)
print(results.getEstimatedParameters())

print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')



import pandas as pd
import biogeme.optimization as opt
import biogeme.biogeme as bio
import biogeme.database as db
import biogeme.models as models
from biogeme.expressions import Beta, Variable
import biogeme.messaging as msg
logger = msg.bioMessage()
logger.setDebug()

df = pd.DataFrame({'Person': [1, 1, 1, 2, 2],
                   'Exclude': [0, 0, 1, 0, 1],
                   'Variable1': [1, 2, 3, 4, 5],
                   'Variable2': [10, 20, 30,40, 50],
                   'Choice': [1, 2, 3, 1, 2],
                   'Av1': [0, 1, 1, 1, 1],
                   'Av2': [1, 1, 1, 1, 1],
                   'Av3': [0, 1, 1, 1, 1]})
myData = db.Database('test', df)

Choice = Variable('Choice')
Variable1 = Variable('Variable1')
Variable2 = Variable('Variable2')
beta1 = Beta('beta1', 0, None, None, 0)
beta2 = Beta('beta2', 0, None, None, 0)
V1 = beta1 * Variable1
V2 = beta2 * Variable2
V3 = 0
V ={1: V1,2: V2,3: V3}

likelihood = models.loglogit(V, av=None, i=Choice)
myBiogeme = bio.BIOGEME(myData, likelihood, numberOfThreads=1)
myBiogeme.modelName = 'simpleExample'
myBiogeme.saveIterations = False
myBiogeme.generateHtml = False
myBiogeme.generatePickle = False
print(myBiogeme)

f, g, h, gdiff, hdiff = myBiogeme.checkDerivatives(verbose=True)

print(gdiff)



"""File 05nestedElasticitiesCI_bootstrap.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 15:57:46 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.

 We calculate disaggregate and aggregate direct arc elasticities, and
 the confidence intervals. The difference with
 05nestedElasticitiesConfidenceIntervals is that the simulated betas
 used to calculated the confidence intervals are drawn from a normal
 distribution based on the bootstrap estimate of the variance
 covariance matrix.
"""

import sys
import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.results as res
from biogeme.expressions import Beta

# Calculate confidence intervals for elasticities requires interval arithmetics
# Use 'pip install pyinterval' if not available on your system.
# Warning: other types of interval packages are also available.
try:
    import interval as ia
except ModuleNotFoundError:
    print('Use "pip install pyinterval" to install a requested package')
    sys.exit()

# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Normalize the weights
sumWeight = database.data['Weight'].sum()
numberOfRows = database.data.shape[0]
normalizedWeight = Weight * numberOfRows / sumWeight

# Calculate the number of accurences of a value in the database
numberOfMales = database.count('Gender', 1)
print(f'Number of males:   {numberOfMales}')

numberOfFemales = database.count('Gender', 2)
print(f'Number of females: {numberOfFemales}')

# For more complex conditions, using directly Pandas
unreportedGender = database.data[
    (database.data['Gender'] != 1) & (database.data['Gender'] != 2)
].count()['Gender']
print(f'Unreported gender: {unreportedGender}')

# List of parameters. Their value will be set later.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME', 0, None, None, 0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER', 0, None, None, 0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE', 0, None, None, 0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE', 0, None, None, 0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED', 0, None, None, 0)
BETA_COST = Beta('BETA_COST', 0, None, None, 0)

# Define new variables. Must be consistent with estimation results.
TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5
male = Gender == 1
female = Gender == 2
unreportedGender = Gender == -1
fulltime = OccupStat == 1
notfulltime = OccupStat != 1

# Definition of utility functions:
V_PT = (
    ASC_PT
    + BETA_TIME_FULLTIME * TimePT_scaled * fulltime
    + BETA_TIME_OTHER * TimePT_scaled * notfulltime
    + BETA_COST * MarginalCostPT_scaled
)
V_CAR = (
    ASC_CAR
    + BETA_TIME_FULLTIME * TimeCar_scaled * fulltime
    + BETA_TIME_OTHER * TimeCar_scaled * notfulltime
    + BETA_COST * CostCarCHF_scaled
)
V_SM = (
    ASC_SM
    + BETA_DIST_MALE * distance_km_scaled * male
    + BETA_DIST_FEMALE * distance_km_scaled * female
    + BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender
)

# Associate utility functions with the numbering of alternatives
V = {0: V_PT, 1: V_CAR, 2: V_SM}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

MU_NOCAR = Beta('MU_NOCAR', 1.0, 1.0, None, 0)
CAR_NEST = 1.0, [1]
NO_CAR_NEST = MU_NOCAR, [0, 2]
nests = CAR_NEST, NO_CAR_NEST

# The choice model is a nested logit
prob_pt = models.nested(V, None, nests, 0)
prob_car = models.nested(V, None, nests, 1)
prob_sm = models.nested(V, None, nests, 2)

# We investigate a scenario where the distance increases by one kilometer.
delta_dist = 1.0
distance_km_scaled_after = (distance_km + delta_dist) / 5

# Utility of the slow mode whem the distance increases by 1 kilometer.
V_SM_after = (
    ASC_SM
    + BETA_DIST_MALE * distance_km_scaled_after * male
    + BETA_DIST_FEMALE * distance_km_scaled_after * female
    + BETA_DIST_UNREPORTED * distance_km_scaled_after * unreportedGender
)

# Associate utility functions with the numbering of alternatives
V_after = {0: V_PT, 1: V_CAR, 2: V_SM_after}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

prob_sm_after = models.nested(V_after, None, nests, 2)

direct_elas_sm_dist = (
    (prob_sm_after - prob_sm) * distance_km / (prob_sm * delta_dist)
)

simulate = {
    'weight': normalizedWeight,
    'Prob. slow modes': prob_sm,
    'direct_elas_sm_dist': direct_elas_sm_dist,
}

biogeme = bio.BIOGEME(database, simulate)
biogeme.modelName = '05nestedElasticitiesCI_Bootstrap'

# Read the estimation results from the file
results = res.bioResults(pickleFile='01nestedEstimation.pickle')

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulatedValues = biogeme.simulate(results.getBetaValues())

# We calculate the elasticities
simulatedValues['Weighted prob. slow modes'] = (
    simulatedValues['weight'] * simulatedValues['Prob. slow modes']
)

denominator_sm = simulatedValues['Weighted prob. slow modes'].sum()

direct_elas_sm_dist = (
    simulatedValues['Weighted prob. slow modes']
    * simulatedValues['direct_elas_sm_dist']
    / denominator_sm
).sum()
print(
    f'Aggregate direct elasticity of slow modes wrt distance: '
    f'{direct_elas_sm_dist:.7f}'
)


print('Calculating confidence interval...')

# Calculate confidence intervals
# Draw values of betas from the distribution of the estimator. The
# bootstrap estimate of the variance-covariance matrix is used.

size = 100

# All betas are simulated
simulatedBetas = np.random.multivariate_normal(
    results.data.betaValues, results.data.bootstrap_varCovar, size
)

# Only those necessary for the simulation are selected.
# The values are also associated with their name
index = [results.data.betaNames.index(b) for b in biogeme.freeBetaNames]
b = [
    {biogeme.freeBetaNames[i]: value for i, value in enumerate(row)}
    for row in simulatedBetas[:, index]
]

# Returns data frame containing, for each simulated value, the left
# and right bounds of the confidence interval calculated by
# simulation.
left, right = biogeme.confidenceIntervals(b, 0.9)

left['Weighted prob. slow modes'] = left['weight'] * left['Prob. slow modes']
right['Weighted prob. slow modes'] = (
    right['weight'] * right['Prob. slow modes']
)
denominator_left = left['Weighted prob. slow modes'].sum()
denominator_right = right['Weighted prob. slow modes'].sum()

# Build an interval object for the denominator
denominator_interval = ia.interval[(denominator_left, denominator_right)]

# Build a list of interval objects, one for each disaggregate elasticity
elas_interval = [
    ia.interval([l, r])
    for l, r in zip(left['direct_elas_sm_dist'], right['direct_elas_sm_dist'])
]

# Build a list of interval objects, one for each term of the numerator
numerator_interval = [
    ia.interval([l, r])
    for l, r in zip(
        left['Weighted prob. slow modes'], right['Weighted prob. slow modes']
    )
]

# Build a list of interval objects, one for each term of the sum
terms_of_the_sum_interval = [
    e * wp / denominator_interval
    for e, wp in zip(elas_interval, numerator_interval)
]

# The interval package apparently does not provide a tool to sum a
# list of intervals. We do it manually. Note that the object interval
# contains a list of ranges (to allow the modeling of disconnected
# intervals). The first of these ranges is x[0], and we access the inf
# and sup values of this range.
sum_interval_left = sum(x[0].inf for x in terms_of_the_sum_interval)
sum_interval_right = sum(x[0].sup for x in terms_of_the_sum_interval)

print(f'[{sum_interval_left}, {sum_interval_right}]')



"""Singleton managing the various levels of messages

:author: Michel Bierlaire
:date: Mon Jul 22 16:12:00 2019

"""

# Too constraining
# pylint: disable=invalid-name,

import datetime
import biogeme.filenames as bf
import biogeme.version as bv
from biogeme.singleton import Singleton


class bioMessage(metaclass=Singleton):
    """Manages the Biogeme messages"""

    def __init__(self, screenLevel=0):
        """Constructor

        :param screenLevel: level of message that must be displayed on the
            screen:

            - 0: no output (default)
            - 1: warnings only
            - 2: only warnings and general information
            - 3: more verbose
            - 4: debug messages


        :type screenLevel: int
        """
        self.screenLevel = screenLevel #: screen verbosity level
        self.types = {
            0: 'Silent',
            1: 'Warning',
            2: 'General',
            3: 'Detailed',
            4: 'Debug',
        } #: names of verbosity levels

        self.resetMessages()

        self.lastLevel = None #: last level used

    def resetMessages(self):
        """Erase all messages"""
        self.messages = []

    def allMessages(self, screenLevel=None):
        """Report all the messages up to a given level.

        :param screenLevel: level of message that must be reported in the file:

           - 0: no output
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose (default)
           - 4: debug messages

             If None (default), all messages are reported.

        :type screenLevel: int

        :return: all messages.
        :rtype: str.
        """
        result = (
            f'*** Messages from biogeme {bv.getVersion()} [{bv.versionDate}]\n'
        )
        for m, i in self.messages:
            if screenLevel is None or i <= screenLevel:
                result += f'{m}\n'
        return result

    def createLog(self, fileLevel=None, fileName='_biogeme'):
        """Creates a log file

        :param fileLevel: level of message that must be reported in the file:

           - 0: no output
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose (default)
           - 4: debug messages

             If None (default), all messages are reported.
        :type fileLevel: int

        :param fileName: name of the file (without extension).
                         Default: '_biogeme'. A file called _biogeme.log
                         will be created.
        :type fileName: string

        :return: name of the file
        :rtype: str
        """
        completeFileName = bf.getNewFileName(fileName, 'log')
        with open(completeFileName, 'w') as f:
            self.general(f'Log file created: {completeFileName}')
            print(f'*** File created {datetime.datetime.now()} ***', file=f)
            print(
                f'*** Log file from biogeme {bv.getVersion()}'
                f' [{bv.versionDate}]',
                file=f,
            )
            for m, i in self.messages:
                if fileLevel is None or i <= fileLevel:
                    print(m, file=f)
        return completeFileName

    def temporarySilence(self):
        """Temporarily turns off the message, remembering the current
        screen level.
        """
        self.lastLevel = self.screenLevel
        self.screenLevel = 0

    def resume(self):
        """Resume the regular operations of the logger after the use of
        temporarySilence
        """
        if self.lastLevel is not None:
            self.screenLevel = self.lastLevel

    def setSilent(self):
        """Set both screen and file levels to 0"""
        self.screenLevel = 0

    def setWarning(self):
        """Set both screen and file levels to 1"""
        self.screenLevel = 1

    def setGeneral(self):
        """Set both screen and file levels to 2"""
        self.screenLevel = 2

    def setDetailed(self):
        """Set both screen and file levels to 3"""
        self.screenLevel = 3

    def setDebug(self):
        """Set both screen and file levels to 4"""
        self.screenLevel = 4

    def setScreenLevel(self, level):
        """Change the level of messaging for the screen

        :param level: level of message that must be displayed on the screen:

           - 0: no output
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose
           - 4: debug messages


        :type level: int
        """
        self.screenLevel = level

    def addMessage(self, text, level):
        """Add a message

        :param text: text of the message.
        :type text: string

        :param level: level of the message

           - 1: warning
           - 2: general information
           - 3: detailed information
           - 4: debug message

        :type level: int

        :note: adding a message of level 0 is meaningless, as it correspond to
            silentmode.
        """
        theLevel = f'< {self.types[level]} >'
        theMessage = (
            f'[{datetime.datetime.now().strftime("%H:%M:%S")}] '
            f'{theLevel:13} {text}'
        )
        if level != 0:
            self.messages.append((theMessage, level))
            if level <= self.screenLevel:
                print(theMessage)

    def warning(self, text):
        """Add a warning

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text, 1)

    def general(self, text):
        """Add a general message

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text, 2)

    def detailed(self, text):
        """Add a detailed message

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text, 3)

    def debug(self, text):
        """Add a debug message

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text, 4)



"""File optima.py

:author: Michel Bierlaire, EPFL
:date: Thu Dec 24 16:51:28 2020

Assisted specification for the Optima case study
"""

# Too constraining
# pylint: disable=invalid-name, undefined-variable

import pandas as pd
import biogeme.database as db
from biogeme import models
import biogeme.messaging as msg
from biogeme import vns
from biogeme import assisted
from biogeme.expressions import (
    Beta,
    log,
    DefineVariable,
    Elem,
    Numeric,
    Variable,
)

logger = msg.bioMessage()
logger.setDebug()

# Read the data
df = pd.read_csv('optima.dat', sep='\t')

df.loc[df['OccupStat'] > 2, 'OccupStat'] = 3
df.loc[df['OccupStat'] == -1, 'OccupStat'] = 3

df.loc[df['Education'] <= 3, 'Education'] = 3
df.loc[df['Education'] <= 3, 'Education'] = 3
df.loc[df['Education'] == 5, 'Education'] = 4
df.loc[df['Education'] == 8, 'Education'] = 7

df.loc[df['TripPurpose'] != 1, 'TripPurpose'] = 2

df.loc[df['CarAvail'] != 3, 'CarAvail'] = 1

database = db.Database('optima', df)

globals().update(database.variables)

exclude = (
    (Choice == -1) + (CostCarCHF < 0) + (CarAvail == 3) * (Choice == 1)
) > 0
database.remove(exclude)

# Definition of new variables


otherSubscription = DefineVariable(
    'otherSubscription',
    ((HalfFareST == 1) + (LineRelST == 1) + (AreaRelST == 1) + (OtherST) == 1)
    > 0,
    database,
)

subscription = DefineVariable(
    'subscription',
    (GenAbST == 1) * 1 + (GenAbST != 1) * otherSubscription * 2,
    database,
)

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5

# Definition of potential nonlinear transforms of variables
def mylog(x):
    """Log of the variable"""
    return 'log', Elem({0: log(x), 1: Numeric(0)}, x == 0)


def sqrt(x):
    """Sqrt of the variable"""
    return 'sqrt', x ** 0.5


def square(x):
    """Square of the variable"""
    return 'square', x ** 2


def boxcox(x, name):
    """Box-Cox transform of the variable"""
    ell = Beta(f'lambda_{name}', 1, 0.0001, 3.0, 0)
    return f'Box-Cox_{name}', models.boxcox(x, ell)


def boxcox_time(x):
    """Box-Cox transform of the variable time"""
    return boxcox(x, 'time')


def boxcox_cost(x):
    """Box-Cox transform of the variable cost"""
    return boxcox(x, 'cost')


def distanceInteraction(x):
    """Nonlinea rinteraction with distance"""
    return 'dist. interaction', x * log(1 + Variable('distance_km') / 1000)


# Define all possible segmentations
all_segmentations = {
    'TripPurpose': (TripPurpose, {1: 'work', 2: 'others'}),
    'Urban': (UrbRur, {1: 'rural', 2: 'urban'}),
    'Language': (LangCode, {1: 'French', 2: 'German'}),
    'Income': (
        Income,
        {
            1: '<2500',
            2: '2051_4000',
            3: '4001_6000',
            4: '6001_8000',
            5: '8001_10000',
            6: '>10000',
            -1: 'unknown',
        },
    ),
    'Gender': (Gender, {1: 'male', 2: 'female', -1: 'unkown'}),
    'Occupation': (
        OccupStat,
        {1: 'full_time', 2: 'partial_time', 3: 'others'},
    ),
    'Subscription': (subscription, {0: 'none', 1: 'GA', 2: 'other'}),
    'CarAvail': (CarAvail, {1: 'yes', 3: 'no'}),
    'Education': (
        Education,
        {3: 'vocational', 4: 'high_school', 6: 'higher_edu', 7: 'university'},
    ),
}


# Define segmentations
segmentations = {
    'Seg. cte': all_segmentations,
    'Seg. cost': all_segmentations,
    'Seg. wait': all_segmentations,
    'Seg. time': all_segmentations,
    'Seg. transfers': all_segmentations,
    'Seg. dist': all_segmentations,
}

# Define the attributes of the alternatives
variables = {
    'PT travel time': TimePT_scaled,
    'PT travel cost': MarginalCostPT_scaled,
    'Car travel time': TimeCar_scaled,
    'Car travel cost': CostCarCHF_scaled,
    'Distance': distance_km_scaled,
    'Transfers': NbTransf,
    'PT Waiting time': WaitingTimePT,
}

# Group the attributes. All attributes in the same group will be
# assoaited with the same nonlinear transform, and the same
# segmentation. Attributes in the same group can be generic or
# alternative specific, except if mentioned otherwise
groupsOfVariables = {
    'Travel time': ['PT travel time', 'Car travel time'],
    'Travel cost': ['PT travel cost', 'Car travel cost'],
}

# In this example, no variable must be alternative specific
genericForbiden = None

# In this example, all the variables must be in the model
forceActive = ['Travel time', 'Distance']

# Associate a list of potential nonlinearities with each group of variable
nonlinearSpecs = {
    'Travel time': [distanceInteraction, mylog, sqrt, square, boxcox_time],
    'PT Waiting time': [mylog, sqrt, square, boxcox_time],
    'Travel cost': [mylog, sqrt, square, boxcox_cost],
}

def negativeParameter(val):
    """ Function verifying the negativity of the parameters"""
    return val < 0


# Specification of the utility function. For each term, it is possible
# to define bounds on the coefficient, and to include a function that
# verifies its validity a posteriori.

utility_pt = [
    (None, 'Seg. cte', (None, None), None),
    ('PT travel time', 'Seg. time', (None, 0), negativeParameter),
    ('PT travel cost', 'Seg. cost', (None, 0), negativeParameter),
    ('Transfers', 'Seg. transfers', (None, 0), negativeParameter),
    ('PT Waiting time', 'Seg. wait', (None, 0), negativeParameter),
]


utility_car = [
    (None, 'Seg. cte', (None, None), None),
    ('Car travel time', 'Seg. time', (None, 0), negativeParameter),
    ('Car travel cost', 'Seg. cost', (None, 0), negativeParameter),
]

utility_sm = [('Distance', 'Seg. dist', (None, 0), negativeParameter)]


utilities = {
    0: ('pt', utility_pt),
    1: ('car', utility_car),
    2: ('sm', utility_sm),
}

availabilities = {0: 1, 1: CarAvail != 3, 2: 1}


# We define potential candidates for the choice model.
def logit(V, av, choice):
    """logit model"""
    return models.loglogit(V, av, choice)


def nested1(V, av, choice):
    """Nested logit model: first specification """
    same = Beta('mu_same', 1, 1, None, 0), [0, 1]
    multiple = 1.0, [2]
    nests = same, multiple
    return models.lognested(V, av, nests, choice)


def nested2(V, av, choice):
    """Nested logit model: second specification """
    onestop = Beta('mu_onestop', 1, 1, None, 0), [1, 2]
    nostop = 1.0, [0]
    nests = nostop, onestop
    return models.lognested(V, av, nests, choice)


def cnl1(V, av, choice):
    """Cross nested logit: fixed alphas"""
    mu_same = Beta('mu_same', 1, 1, None, 0)
    mu_onestop = Beta('mu_onestop', 1, 1, None, 0)
    alpha_same = {0: 1.0, 1: 0.5, 2: 0}
    alpha_onestop = {0: 0, 1: 0.5, 2: 1}
    nest_same = mu_same, alpha_same
    nest_onestop = mu_onestop, alpha_onestop
    nests = nest_onestop, nest_same
    return models.logcnl_avail(V, av, nests, choice)


def cnl2(V, av, choice):
    """Cross nested logit: estimated alphas"""
    alpha = Beta('alpha', 0.5, 0, 1, 0)
    mu_same = Beta('mu_same', 1, 1, None, 0)
    mu_onestop = Beta('mu_onestop', 1, 1, None, 0)
    alpha_same = {0: 1.0, 1: alpha, 2: 0}
    alpha_onestop = {0: 0, 1: 1 - alpha, 2: 1}
    nest_same = mu_same, alpha_same
    nest_onestop = mu_onestop, alpha_onestop
    nests = nest_onestop, nest_same
    return models.logcnl_avail(V, av, nests, choice)


# We provide names to these candidates
myModels = {'Logit': logit,
            'Nested 1': nested1,
            'Nested 2': nested2,
            'Cross nested 1': cnl1,
            'Cross nested 2': cnl2}

# Definition of the specification problem, gathering all information
# defined above.
theProblem = assisted.specificationProblem(
    'Optima',
    database,
    variables,
    groupsOfVariables,
    genericForbiden,
    forceActive,
    nonlinearSpecs,
    segmentations,
    utilities,
    availabilities,
    Choice,
    myModels,
)

theProblem.maximumNumberOfParameters = 100

# We propose several specifications to initialize the algorithm.
# For each group of attributes, we decide if it is nonlinear, and generic.
nl1 = {'Travel time': (None, False), 'Distance': (None, False)}

# For each segmentation, we decided which dimensions are active.
sg1 = {
    'Seg. cte': ['Subscription'],
    'Seg. time': [],
    'Seg. cost': [],
    'Seg. dist': [],
}

initSolutions = [theProblem.generateSolution(nl1, sg1, 'Logit')]

# Optimization algorithm
vns.vns(
    theProblem,
    initSolutions,
    archiveInputFile='optimaPareto.pickle',
    pickleOutputFile='optimaPareto.pickle',
)



"""File 05normalMixture.py

:author: Michel Bierlaire, EPFL
:date: Sat Sep  7 18:23:01 2019

 Example of a mixture of logit models, using Monte-Carlo integration.
 Three alternatives: Train, Car and Swissmetro
 SP data
"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import Beta, DefineVariable, bioDraws, log, MonteCarlo

# Read the data
df = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to investigate the database. For example:
# print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
# remove = (((database.data.PURPOSE != 1) &
#           (database.data.PURPOSE != 3)) |
#          (database.data.CHOICE == 0))
# database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# Definition of new variables: adding columns to the database
CAR_AV_SP = DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0), database)
TRAIN_AV_SP = DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0), database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0, database)
TRAIN_COST_SCALED = DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100, database
)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0, database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100, database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100, database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100, database)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(prob))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# These notes will be included as such in the report file.
userNotes = (
    'Example of a mixture of logit models with three alternatives, '
    'approximated using Monte-Carlo integration.'
)

# Create the Biogeme object
biogeme = bio.BIOGEME(
    database, logprob, numberOfDraws=100000, userNotes=userNotes
)
biogeme.modelName = '05normalMixture'

# Estimate the parameters
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)



"""File 03scale.py

:author: Michel Bierlaire, EPFL
:date: Thu Sep  6 15:14:39 2018

 Illustrates heteroscedastic specification. A different scale is
 associated with different segments of the sample.
 Three alternatives: Train, Car and Swissmetro
 SP data

"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import Beta, DefineVariable

# Read the data
df = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to investigate the database. For example:
# print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
# remove = (((database.data.PURPOSE != 1) &
#           (database.data.PURPOSE != 3)) |
#          (database.data.CHOICE == 0))
# database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
Scale_group3 = Beta('Scale_group3', 1, 0.001, None, 0)

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# Definition of new variables: adding columns to the database
CAR_AV_SP = DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0), database)
TRAIN_AV_SP = DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0), database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0, database)
TRAIN_COST_SCALED = DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100, database
)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0, database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100, database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100, database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100, database)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Scale associated with group 3 is estimated
scale = (GROUP != 3) + (GROUP == 3) * Scale_group3

# Scale the utility functions, and associate them with the numbering
# of alternatives
V = {1: scale * V1, 2: scale * V2, 3: scale * V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Define level of verbosity
logger = msg.bioMessage()
logger.setSilent()
# logger.setWarning()
# logger.setGeneral()
# logger.setDetailed()

# These notes will be included as such in the report file.
userNotes = (
    'Illustrates heteroscedastic specification. A different scale is'
    ' associated with different segments of the sample.'
)

# Create the Biogeme object
biogeme = bio.BIOGEME(database, logprob, userNotes=userNotes)
biogeme.modelName = '03scale'

# Estimate the parameters
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)



"""File 03choiceOnly_mc.py

Choice model with the latent variable.
Mixture of logit, using Monte-Carlo integration
No measurement equation for the indicators.

:author: Michel Bierlaire, EPFL
:date: Sat May 30 18:16:04 2020

"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from  biogeme import models
import biogeme.optimization as opt
import biogeme.messaging as msg
from biogeme.expressions import (
    Beta,
    DefineVariable,
    bioDraws,
    MonteCarlo,
    exp,
    log,
)

# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)


# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

### Variables

# Piecewise linear definition of income
ScaledIncome = DefineVariable(
    'ScaledIncome', CalculatedIncome / 1000, database
)

thresholds = [None, 4, 6, 8, 10, None]
formulaIncome = models.piecewiseFormula(
    ScaledIncome, thresholds, [0.0, 0.0, 0.0, 0.0, 0.0]
)

# Definition of other variables
age_65_more = DefineVariable('age_65_more', age >= 65, database)
moreThanOneCar = DefineVariable('moreThanOneCar', NbCar > 1, database)
moreThanOneBike = DefineVariable('moreThanOneBike', NbBicy > 1, database)
individualHouse = DefineVariable('individualHouse', HouseType == 1, database)
male = DefineVariable('male', Gender == 1, database)
haveChildren = DefineVariable(
    'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0, database
)
haveGA = DefineVariable('haveGA', GenAbST == 1, database)
highEducation = DefineVariable('highEducation', Education >= 6, database)

# Parameters to be estimated
coef_intercept = Beta('coef_intercept', 0.0, None, None, 1)
coef_age_65_more = Beta('coef_age_65_more', 0.0, None, None, 0)
coef_haveGA = Beta('coef_haveGA', 0.0, None, None, 0)
coef_moreThanOneCar = Beta('coef_moreThanOneCar', 0.0, None, None, 0)
coef_moreThanOneBike = Beta('coef_moreThanOneBike', 0.0, None, None, 0)
coef_individualHouse = Beta('coef_individualHouse', 0.0, None, None, 0)
coef_male = Beta('coef_male', 0.0, None, None, 0)
coef_haveChildren = Beta('coef_haveChildren', 0.0, None, None, 0)
coef_highEducation = Beta('coef_highEducation', 0.0, None, None, 0)

### Latent variable: structural equation

sigma_s = Beta('sigma_s', 1, None, None, 0)

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formulaIncome
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + sigma_s * bioDraws('EC', 'NORMAL_MLHS')
)

# Choice model

ASC_CAR = Beta('ASC_CAR', 0.0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0.0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0.0, None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', 0.0, None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', 0.0, None, None, 0)
BETA_DIST = Beta('BETA_DIST', 0.0, None, None, 0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', 0.0, None, 0, 0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL', 0.0, None, None, 0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', 0.0, None, 0, 0)
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', -1.0, None, None, 0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', 0.0, None, None, 0)

TimePT_scaled = DefineVariable('TimePT_scaled', TimePT / 200, database)
TimeCar_scaled = DefineVariable('TimeCar_scaled', TimeCar / 200, database)
MarginalCostPT_scaled = DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10, database
)
CostCarCHF_scaled = DefineVariable(
    'CostCarCHF_scaled', CostCarCHF / 10, database
)
distance_km_scaled = DefineVariable(
    'distance_km_scaled', distance_km / 5, database
)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1, database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1, database)

### Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel)
condprob = models.logit(V, None, Choice)
# We integrate over omega using numerical integration
loglike = log(MonteCarlo(condprob))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, loglike, numberOfDraws=10000)
biogeme.modelName = '03choiceOnly_mc'

# Estimate the parameters
results = biogeme.estimate(algorithm=opt.bioNewton)

print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')
results.writeLaTeX()
print(f'LaTeX file: {results.data.latexFileName}')



"""File airline.py

:author: Michel Bierlaire, EPFL
:date: Mon Dec 21 15:24:50 2020

Assisted specification for the airline cases tudy
"""

# Too constraining
# pylint: disable=invalid-name, undefined-variable

import pandas as pd
import biogeme.database as db
import biogeme.models as models
import biogeme.messaging as msg
import biogeme.vns as vns
import biogeme.assisted as assisted
from biogeme.expressions import (
    Beta,
    log,
    DefineVariable,
    Elem,
    Numeric,
    Variable,
)

logger = msg.bioMessage()
logger.setDebug()

# Read the data
df = pd.read_csv('airline.dat', sep='\t')

# Update some data
df.loc[df['q17_Gender'] == 99, 'q17_Gender'] = -1
df.loc[df['q20_Education'] == 99, 'q20_Education'] = -1

database = db.Database('airline', df)
globals().update(database.variables)

exclude = ArrivalTimeHours_1 == -1
database.remove(exclude)

# Definition of new variables

chosenAlternative = (
    (BestAlternative_1 * 1) + (BestAlternative_2 * 2) + (BestAlternative_3 * 3)
)

DepartureTimeSensitive = DefineVariable(
    'DepartureTimeSensitive', q11_DepartureOrArrivalIsImportant == 1, database
)
ArrivalTimeSensitive = DefineVariable(
    'ArrivalTimeSensitive', q11_DepartureOrArrivalIsImportant == 2, database
)
Missing = DefineVariable(
    'Missing',
    (q11_DepartureOrArrivalIsImportant != 1)
    * (q11_DepartureOrArrivalIsImportant != 2),
    database,
)

DesiredDepartureTime = DefineVariable(
    'DesiredDepartureTime', q12_IdealDepTime, database
)
DesiredArrivalTime = DefineVariable(
    'DesiredArrivalTime', q13_IdealArrTime, database
)
ScheduledDelay_1 = DefineVariable(
    'ScheduledDelay_1',
    (DepartureTimeSensitive * (DepartureTimeMins_1 - DesiredDepartureTime))
    + (ArrivalTimeSensitive * (ArrivalTimeMins_1 - DesiredArrivalTime)),
    database,
)

ScheduledDelay_2 = DefineVariable(
    'ScheduledDelay_2',
    (DepartureTimeSensitive * (DepartureTimeMins_2 - DesiredDepartureTime))
    + (ArrivalTimeSensitive * (ArrivalTimeMins_2 - DesiredArrivalTime)),
    database,
)

ScheduledDelay_3 = DefineVariable(
    'ScheduledDelay_3',
    (DepartureTimeSensitive * (DepartureTimeMins_3 - DesiredDepartureTime))
    + (ArrivalTimeSensitive * (ArrivalTimeMins_3 - DesiredArrivalTime)),
    database,
)

Opt1_SchedDelayEarly = DefineVariable(
    'Opt1_SchedDelayEarly',
    (-(ScheduledDelay_1) * (ScheduledDelay_1 < 0)) / 60,
    database,
)
Opt2_SchedDelayEarly = DefineVariable(
    'Opt2_SchedDelayEarly',
    (-(ScheduledDelay_2) * (ScheduledDelay_2 < 0)) / 60,
    database,
)
Opt3_SchedDelayEarly = DefineVariable(
    'Opt3_SchedDelayEarly',
    (-(ScheduledDelay_3) * (ScheduledDelay_3 < 0)) / 60,
    database,
)

Opt1_SchedDelayLate = DefineVariable(
    'Opt1_SchedDelayLate',
    (ScheduledDelay_1 * (ScheduledDelay_1 > 0)) / 60,
    database,
)
Opt2_SchedDelayLate = DefineVariable(
    'Opt2_SchedDelayLate',
    (ScheduledDelay_2 * (ScheduledDelay_2 > 0)) / 60,
    database,
)
Opt3_SchedDelayLate = DefineVariable(
    'Opt3_SchedDelayLate',
    (ScheduledDelay_3 * (ScheduledDelay_3 > 0)) / 60,
    database,
)


# Definition of potential nonlinear transforms of variables
def incomeInteraction(x):
    """Defines an interaction with income"""
    return 'inc. interaction', x / Variable('Cont_Income')


def incomeInteraction2(x):
    """Defines another interaction with income"""
    return 'inc. interaction 2', x + x / Variable('Cont_Income')


def logincomeInteraction(x):
    """Defines an interaction with between the log and income"""
    return 'inc. interaction', log(x) / Variable('Cont_Income')


def sqrtincomeInteraction(x):
    """Defines an interaction with between the sqrt and income"""
    return 'inc. interaction', x ** 0.5 / Variable('Cont_Income')


def mylog(x):
    """Log of the variable"""
    return 'log', Elem({0: log(x), 1: Numeric(0)}, x == 0)


def sqrt(x):
    """Sqrt of the variable"""
    return 'sqrt', x ** 0.5


def square(x):
    """Square of the variable"""
    return 'square', x ** 2


def piecewise(x, thresholds, name):
    """Piecewise linear specification"""
    piecewiseVariables = models.piecewiseVariables(x, thresholds)
    formula = piecewiseVariables[0]
    for k in range(1, len(thresholds) - 1):
        formula += (
            Beta(
                f'pw_{name}_{thresholds[k-1]}_{thresholds[k]}',
                0,
                None,
                None,
                0,
            )
            * piecewiseVariables[k]
        )
    return (f'piecewise_{thresholds}', formula)


def piecewise_time_2(x):
    """Piecewise linear for time :math:`0, 2, 8, +\\infty`"""
    return piecewise(x, [0, 2, 8, None], 'time')


def piecewise_time_1(x):
    """Piecewise linear for time :math:`0, 2, 4, 6, 8, +\\infty`"""
    return piecewise(x, [0, 2, 4, 6, 8, None], 'time')


def boxcox(x, name):
    """Box-Cox transform of the variable"""
    ell = Beta(f'lambda_{name}', 1, 0.0001, 3.0, 0)
    return f'Box-Cox_{name}', models.boxcox(x, ell)


def boxcox_time(x):
    """Box-Cox transform of the variable time"""
    return boxcox(x, 'time')


def boxcox_fare(x):
    """Box-Cox transform of the variable fare"""
    return boxcox(x, 'fare')


# Define all possible segmentations
all_segmentations = {
    'TripPurpose': (
        q02_TripPurpose,
        {
            1: 'business',
            2: 'leisure',
            3: 'attending conf.',
            4: 'business & leisure',
            0: 'unknown',
        },
    ),
    'Gender': (q17_Gender, {1: 'male', 2: 'female', -1: 'unknown'}),
    'Education': (
        q20_Education,
        {
            1: 'less than high school',
            2: 'high school',
            3: 'some college',
            4: 'associate occ.',
            5: 'associate acad.',
            6: 'bachelor',
            7: 'master',
            8: 'professional',
            9: 'doctorate',
            -1: 'unkonown',
        },
    ),
    'Importance': (
        q11_DepartureOrArrivalIsImportant,
        {1: 'departure', 2: 'arrival', 3: 'not important'},
    ),
    'Who pays': (
        q03_WhoPays,
        {1: 'traveler', 2: 'employer', 3: 'third party', 0: 'unknown'},
    ),
}

# Define segmentations
segmentations = {
    'Seg. cte': all_segmentations,
    'Seg. fare': all_segmentations,
    'Seg. time': all_segmentations,
    'Seg. delay': all_segmentations,
    'Seg. legroom': all_segmentations,
}

# Define the attributes of the alternatives
variables = {
    'Fare direct': Fare_1,
    'Fare same': Fare_2,
    'Fare multiple': Fare_3,
    'Legroom direct': Legroom_1,
    'Legroom same': Legroom_2,
    'Legroom multiple': Legroom_3,
    'Time direct': TripTimeHours_1,
    'Time same': TripTimeHours_2,
    'Time multiple': TripTimeHours_3,
    'Early direct': Opt1_SchedDelayEarly,
    'Early same': Opt2_SchedDelayEarly,
    'Early multiple': Opt3_SchedDelayEarly,
    'Late direct': Opt1_SchedDelayLate,
    'Late same': Opt2_SchedDelayLate,
    'Late multiple': Opt3_SchedDelayLate,
}


# Group the attributes. All attributes in the same group will be
# assoaited with the same nonlinear transform, and the same
# segmentation. Attributes in the same group can be generic or
# alternative specific, except if mentioned otherwise
groupsOfVariables = {
    'Fare': ['Fare direct', 'Fare same', 'Fare multiple'],
    'Legroom': ['Legroom direct', 'Legroom same', 'Legroom multiple'],
    'Time': ['Time direct', 'Time same', 'Time multiple'],
    'Early': ['Early direct', 'Early same', 'Early multiple'],
    'Late': ['Late direct', 'Late same', 'Late multiple'],
}

# In this example, no variable must be alternative specific
genericForbiden = None

# In this example, all the variables must be in the model
forceActive = list(groupsOfVariables.keys())

# Associate a list of potential nonlinearities with each group of variable
nonlinearSpecs = {
    'Time': [
        mylog,
        sqrt,
        square,
        boxcox_time,
        piecewise_time_1,
        piecewise_time_2,
    ],
    'Fare': [
        incomeInteraction,
        incomeInteraction2,
        logincomeInteraction,
        sqrtincomeInteraction,
        mylog,
        sqrt,
        square,
        boxcox_fare,
    ],
    'Legroom': [mylog, sqrt, square],
    'Early': [mylog, sqrt, square],
    'Late': [mylog, sqrt, square],
}

# Specification of the utility function. For each term, it is possible
# to define bounds on the coefficient, and to include a function that
# verifies its validity a posteriori.

utility_direct = [
    ('Fare direct', 'Seg. fare', (None, 0), None),
    ('Legroom direct', 'Seg. legroom', (0, None), None),
    ('Early direct', 'Seg. delay', (None, 0), None),
    ('Late direct', 'Seg. delay', (None, 0), None),
    ('Time direct', 'Seg. time', (None, 0), None),
]

utility_same = [
    (None, 'Seg. cte', (None, None), None),
    ('Fare same', 'Seg. fare', (None, 0), None),
    ('Legroom same', 'Seg. legroom', (0, None), None),
    ('Early same', 'Seg. delay', (None, 0), None),
    ('Late same', 'Seg. delay', (None, 0), None),
    ('Time same', 'Seg. time', (None, 0), None),
]

utility_multiple = [
    (None, 'Seg. cte', (None, None), None),
    ('Fare multiple', 'Seg. fare', (None, 0), None),
    ('Legroom multiple', 'Seg. legroom', (0, None), None),
    ('Early multiple', 'Seg. delay', (None, 0), None),
    ('Late multiple', 'Seg. delay', (None, 0), None),
    ('Time multiple', 'Seg. time', (None, 0), None),
]

utilities = {
    1: ('Non stop', utility_direct),
    2: ('Same airline', utility_same),
    3: ('Multiple airlines', utility_multiple),
}

availabilities = {1: 1, 2: 1, 3: 1}


# We define potential candidates for the choice model.
def logit(V, av, choice):
    """logit model"""
    return models.loglogit(V, av, choice)


def nested1(V, av, choice):
    """Nested logit model: no stop / one stop"""
    onestop = Beta('mu_onestop', 1, 1, None, 0), [2, 3]
    nonstop = 1.0, [1]
    nests = nonstop, onestop
    return models.lognested(V, av, nests, choice)


def nested2(V, av, choice):
    """Nested logit model: same / multiple"""
    same = Beta('mu_same', 1, 1, None, 0), [1, 2]
    multiple = 1.0, [3]
    nests = same, multiple
    return models.lognested(V, av, nests, choice)


def cnl1(V, av, choice):
    """Cross nested logit: fixed alphas"""
    mu_onestop = Beta('mu_onestop', 1, 1, None, 0)
    mu_same = Beta('mu_same', 1, 1, None, 0)
    alpha_onestop = {1: 1.0, 2: 0.5, 3: 1}
    alpha_same = {1: 1.0, 2: 0.5, 3: 1}
    nest_onestop = mu_onestop, alpha_onestop
    nest_same = mu_same, alpha_same
    nests = nest_onestop, nest_same
    return models.logcnl_avail(V, av, nests, choice)


def cnl2(V, av, choice):
    """Cross nested logit: estimated alphas"""
    alpha = Beta('alpha', 0.5, 0, 1, 0)
    mu_onestop = Beta('mu_onestop', 1, 1, None, 0)
    mu_same = Beta('mu_same', 1, 1, None, 0)
    alpha_onestop = {1: 1.0, 2: alpha, 3: 1}
    alpha_same = {1: 1.0, 2: 1 - alpha, 3: 1}
    nest_onestop = mu_onestop, alpha_onestop
    nest_same = mu_same, alpha_same
    nests = nest_onestop, nest_same
    return models.logcnl_avail(V, av, nests, choice)


# We provide names to these candidates
myModels = {
    'Logit': logit,
    'Nested one stop': nested1,
    'Nested same': nested2,
    'CNL alpha fixed': cnl1,
    'CNL alpha est.': cnl2,
}

# Definition of the specification problem, gathering all information defined above.
theProblem = assisted.specificationProblem(
    'Airline',
    database,
    variables,
    groupsOfVariables,
    genericForbiden,
    forceActive,
    nonlinearSpecs,
    segmentations,
    utilities,
    availabilities,
    chosenAlternative,
    myModels,
)

theProblem.maximumNumberOfParameters = 300

# We propose several specifications to initialize the algorithm.
# For each group of attributes, we decide if it is nonlinear, and generic.
nl1 = {
    'Time': (None, False),
    'Fare': (None, False),
    'Legroom': (None, False),
    'Early': (None, False),
    'Late': (None, False),
}

nl2 = {
    'Time': (5, False),
    'Fare': (0, False),
    'Legroom': (None, False),
    'Early': (None, False),
    'Late': (None, False),
}

nl3 = {
    'Time': (5, False),
    'Fare': (1, False),
    'Legroom': (None, False),
    'Early': (None, False),
    'Late': (None, False),
}

# For each segmentation, we decided which dimensions are active.
sg1 = {'Seg. cte': [], 'Seg. legroom': ['TripPurpose', 'Gender']}

sg2 = {'Seg. cte': ['TripPurpose'], 'Seg. legroom': ['Gender']}


initSolutions = [
    theProblem.generateSolution(nl2, sg2, 'Logit'),
    theProblem.generateSolution(nl2, sg2, 'Nested one stop'),
    theProblem.generateSolution(nl2, sg2, 'Nested same'),
    theProblem.generateSolution(nl2, sg2, 'CNL alpha fixed'),
    theProblem.generateSolution(nl2, sg2, 'CNL alpha est.'),
]

# Optimization algorithm
vns.vns(
    theProblem,
    initSolutions,
    archiveInputFile='airlinePareto.pickle',
    pickleOutputFile='airlinePareto.pickle',
)



"""File: 07estimationMonteCarlo_mlhs_500.py

 Author: Michel Bierlaire, EPFL
 Date: Wed Dec 11 17:24:00 2019

Estimation of a mixtures of logit models where the integral is
approximated using MonteCarlo integration, with Modified
Latin Hypercube Sampling draws.

"""

# pylint: disable=invalid-name, undefined-variable

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, bioDraws, MonteCarlo, log

pandas = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', pandas)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_MLHS')

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)
CAR_AV_SP = CAR_AV * (SP != 0)
TRAIN_AV_SP = TRAIN_AV * (SP != 0)
TRAIN_TT_SCALED = TRAIN_TT / 100.0
TRAIN_COST_SCALED = TRAIN_COST / 100
SM_TT_SCALED = SM_TT / 100.0
SM_COST_SCALED = SM_COST / 100
CAR_TT_SCALED = CAR_TT / 100
CAR_CO_SCALED = CAR_CO / 100

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
prob = models.logit(V, av, CHOICE)
logprob = log(MonteCarlo(prob))

R = 500
biogeme = bio.BIOGEME(database, logprob, numberOfDraws=R)

biogeme.modelName = '07estimationMonteCarlo_mlhs_500'
results = biogeme.estimate()

# Get the results in a pandas table
pandasResults = results.getEstimatedParameters()
print(pandasResults)



import os

ignoreDirectory = [
    './workingNotToDistribute',
    './assisted',
    './montecarlo',
]

htmlToCheck = [
    './swissmetro/05normalMixture_allAlgos_CFSQP.html',
    './swissmetro/05normalMixture_allAlgos_LS-BFGS.html',
    './swissmetro/05normalMixture_allAlgos_Line search.html',
    './swissmetro/05normalMixture_allAlgos_Simple bounds BFGS fCG.html',
    './swissmetro/05normalMixture_allAlgos_Simple bounds BFGS iCG.html',
    './swissmetro/05normalMixture_allAlgos_Simple bounds Newton fCG.html',
    './swissmetro/05normalMixture_allAlgos_Simple bounds Newton iCG.html',
    './swissmetro/05normalMixture_allAlgos_Simple bounds hybrid fCG.html',
    './swissmetro/05normalMixture_allAlgos_Simple bounds hybrid iCG.html',
    './swissmetro/05normalMixture_allAlgos_TR-BFGS.html',
    './swissmetro/05normalMixture_allAlgos_Trust region (cg).html',
    './swissmetro/05normalMixture_allAlgos_Trust region (dogleg).html',
    './swissmetro/05normalMixture_allAlgos_scipy.html',
    './swissmetro/01logit_allAlgos_LS-BFGS.html',
    './swissmetro/01logit_allAlgos_Line search.html',
    './swissmetro/01logit_allAlgos_Simple bounds BFGS.html',
    './swissmetro/01logit_allAlgos_Simple bounds Newton.html',
    './swissmetro/01logit_allAlgos_Simple bounds hybrid.html',
    './swissmetro/01logit_allAlgos_TR-BFGS.html',
    './swissmetro/01logit_allAlgos_Trust region (cg).html',
    './swissmetro/01logit_allAlgos_Trust region (dogleg).html',
    './swissmetro/01logit_allAlgos_scipy.html',
    './swissmetro/09nested_allAlgos_Simple bounds BFGS.html',
    './swissmetro/09nested_allAlgos_Simple bounds Newton.html',
    './swissmetro/09nested_allAlgos_Simple bounds hybrid 20%.html',
    './swissmetro/09nested_allAlgos_Simple bounds hybrid 50%.html',
    './swissmetro/09nested_allAlgos_Simple bounds hybrid 80%.html',
    './swissmetro/09nested_allAlgos_scipy.html',
]
htmlNotNeed = [
    './vns/runknapsack.html',
    './vns/knapsack.html',
    './indicators/03nestedElasticities.html',
    './indicators/05nestedElasticitiesConfidenceIntervals.html',
    './indicators/05nestedElasticitiesCI_Bootstrap.html',
    './indicators/02nestedPlot.html',
    './indicators/04nestedElasticities.html',
    './indicators/05nestedElasticities.html',
    './indicators/02nestedSimulation.html',
    './indicators/06nestedWTP.html',
    './swissmetro/01logit_simul.html',
    './swissmetro/05normalMixture_simul.html',
    './swissmetro/01logit_allAlgos.html',
    './swissmetro/11cnl_simul.html',
    './swissmetro/19individualLevelParameters.html',
    './swissmetro/09nested_allAlgos.html',
    './swissmetro/05normalMixture_allAlgos.html',
    './latent/00factorAnalysis.html',
]

version = '3.2.7'

def checkFile(theFile):
    if theFile.is_file() and theFile.name.endswith('py'):
        current_file = os.path.split(theFile.path)
        dir = current_file[0]
        if dir not in ignoreDirectory:
            root = theFile.name.split(".")[0]
            html = f'{dir}/{root}.html'
            if html not in htmlNotNeed:
                if not os.path.isfile(html):
                    print(f'File {html} is missing')
                else:
                    with open(html) as f:
                        content = f.readlines()
                        for line in content:
                            if ('<p>biogeme' in line and
                                not f'biogeme {version}' in line):
                                print(f'Wrong version for {html}: {line}')

with os.scandir('.') as root_dir:
    for path in root_dir:
        if path.is_dir(follow_symlinks=False):
            with os.scandir(path.path) as local:
                if not path.path in ignoreDirectory:
                    print(f'----- {path.path} -----')
                    for file in local:
                        checkFile(file)



"""File m02_sequential_estimation.py

Choice model with the latent variable.
Mixture of logit.
Measurement equation for the indicators.
Sequential estimation.

:author: Michel Bierlaire, EPFL
:date: Tue Jul  6 18:04:50 2021

"""

import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
import biogeme.results as res
import biogeme.messaging as msg
import biogeme.optimization as opt
from biogeme import models
from biogeme.expressions import (
    Beta,
    DefineVariable,
    Variable,
    RandomVariable,
    exp,
    log,
    Integrate,
)

# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# The following statement allows you to use the names of the variable
# as Python variable.
Choice = Variable('Choice')
Gender = Variable('Gender')
FamilSitu = Variable('FamilSitu')
Education = Variable('Education')
ResidChild = Variable('ResidChild')
SocioProfCat = Variable('SocioProfCat')
age = Variable('age')
TimePT = Variable('TimePT')
TimeCar = Variable('TimeCar')
MarginalCostPT = Variable('MarginalCostPT')
CostCarCHF = Variable('CostCarCHF')
distance_km = Variable('distance_km')
TripPurpose = Variable('TripPurpose')
WaitingTimePT = Variable('WaitingTimePT')
Choice = Variable('Choice')

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Read the estimates from the structural equation estimation
NAME = 'm01_latent_variable'
try:
    structResults = res.bioResults(pickleFile=f'{NAME}.pickle')
except FileNotFoundError:
    print(
        f'Run first the script {NAME}.py in order to generate the file '
        f'{NAME}.pickle.'
    )
    sys.exit()
structBetas = structResults.getBetaValues()

### Variables

# Definition of other variables
male = DefineVariable('male', Gender == 1, database)

haveChildren = DefineVariable(
    'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0, database
)

highEducation = DefineVariable('highEducation', Education >= 6, database)

childCenter = DefineVariable(
    'childCenter', ((ResidChild == 1) + (ResidChild == 2)) > 0, database
)

childSuburb = DefineVariable(
    'childSuburb', ((ResidChild == 3) + (ResidChild == 4)) > 0, database
)


### Coefficients

coef_intercept = structBetas['coef_intercept']
coef_age_30_less = structBetas['coef_age_30_less']
coef_male = structBetas['coef_male']
coef_haveChildren = structBetas['coef_haveChildren']
coef_highEducation = structBetas['coef_highEducation']
coef_artisans = structBetas['coef_artisans']
coef_employees = structBetas['coef_employees']
coef_child_center = structBetas['coef_child_center']
coef_child_suburb = structBetas['coef_child_suburb']

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to
# write it across several lines, each line must terminate with
# the \ symbol

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

ACTIVELIFE = (
    coef_intercept
    + coef_child_center * childCenter
    + coef_child_suburb * childSuburb
    + coef_highEducation * highEducation
    + coef_artisans * (SocioProfCat == 5)
    + coef_employees * (SocioProfCat == 6)
    + coef_age_30_less * (age <= 30)
    + coef_male * male
    + coef_haveChildren * haveChildren
    + sigma_s * omega
)

# Choice model
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', 0, None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', 0, None, None, 0)
BETA_DIST = Beta('BETA_DIST', 0, None, None, 0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', 0, None, 0, 0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', 0, None, 0, 0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', 0, None, None, 0)

# The coefficient of the latent variable should be initialized to
# something different from zero. If not, the algorithm may be trapped
# in a local optimum, and never change the value.
BETA_TIME_PT_AL = Beta('BETA_TIME_PT_AL', 1, None, None, 0)
BETA_TIME_CAR_AL = Beta('BETA_TIME_CAR_AL', -1, None, None, 0)


TimePT_scaled = DefineVariable('TimePT_scaled', TimePT / 200, database)
TimeCar_scaled = DefineVariable('TimeCar_scaled', TimeCar / 200, database)
MarginalCostPT_scaled = DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10, database
)
CostCarCHF_scaled = DefineVariable(
    'CostCarCHF_scaled', CostCarCHF / 10, database
)
distance_km_scaled = DefineVariable(
    'distance_km_scaled', distance_km / 5, database
)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1, database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1, database)

### Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_AL * ACTIVELIFE)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_AL * ACTIVELIFE)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel)
condprob = models.logit(V, None, Choice)
# We integrate over omega using numerical integration
loglike = log(Integrate(condprob * density, 'omega'))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, loglike)
biogeme.modelName = 'm02_sequential_estimation'

# Estimate the parameters
results = biogeme.estimate(algorithm=opt.bioBfgs)
print(results.getEstimatedParameters())

print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')



"""File 06nestedWTP.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 14:01:00 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We calculate and plot willingness to pay.
"""

import sys
import matplotlib.pyplot as plt
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.results as res
import biogeme.messaging as msg

from biogeme.expressions import Beta, Derive


logger = msg.bioMessage()
logger.setGeneral()

# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Normalize the weights
sumWeight = database.data['Weight'].sum()
numberOfRows = database.data.shape[0]
normalizedWeight = Weight * numberOfRows / sumWeight

# Calculate the number of accurences of a value in the database
numberOfMales = database.count('Gender', 1)
print(f'Number of males:   {numberOfMales}')
numberOfFemales = database.count('Gender', 2)
print(f'Number of females: {numberOfFemales}')

# For more complex conditions, we use Pandas
unreportedGender = database.data[
    (database.data['Gender'] != 1) & (database.data['Gender'] != 2)
].count()['Gender']
print(f'Unreported gender: {unreportedGender}')


# List of parameters. Their value will be set later.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME', 0, None, None, 0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER', 0, None, None, 0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE', 0, None, None, 0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE', 0, None, None, 0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED', 0, None, None, 0)
BETA_COST = Beta('BETA_COST', 0, None, None, 0)

# Define new variables. Must be consistent with estimation results.
TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5
male = Gender == 1
female = Gender == 2
unreportedGender = Gender == -1
fulltime = OccupStat == 1
notfulltime = OccupStat != 1

# Definition of utility functions:
V_PT = (
    ASC_PT
    + BETA_TIME_FULLTIME * TimePT_scaled * fulltime
    + BETA_TIME_OTHER * TimePT_scaled * notfulltime
    + BETA_COST * MarginalCostPT_scaled
)
V_CAR = (
    ASC_CAR
    + BETA_TIME_FULLTIME * TimeCar_scaled * fulltime
    + BETA_TIME_OTHER * TimeCar_scaled * notfulltime
    + BETA_COST * CostCarCHF_scaled
)
V_SM = (
    ASC_SM
    + BETA_DIST_MALE * distance_km_scaled * male
    + BETA_DIST_FEMALE * distance_km_scaled * female
    + BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender
)

WTP_PT_TIME = Derive(V_PT, 'TimePT') / Derive(V_PT, 'MarginalCostPT')
WTP_CAR_TIME = Derive(V_CAR, 'TimeCar') / Derive(V_CAR, 'CostCarCHF')

simulate = {
    'weight': normalizedWeight,
    'WTP PT time': WTP_PT_TIME,
    'WTP CAR time': WTP_CAR_TIME,
}

biogeme = bio.BIOGEME(database, simulate, removeUnusedVariables=False)
biogeme.modelName = '06nestedWTP'

# Read the estimation results from the file.
try:
    results = res.bioResults(pickleFile='01nestedEstimation.pickle')
except FileNotFoundError:
    sys.exit(
        'Run first the script 01nestedEstimation.py in order to generate '
        'the file 01nestedEstimation.pickle.'
    )

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulatedValues = biogeme.simulate(results.getBetaValues())

wtpcar = (
    60 * simulatedValues['WTP CAR time'] * simulatedValues['weight']
).mean()

# Calculate confidence intervals
b = results.getBetasForSensitivityAnalysis(biogeme.freeBetaNames, size=100)

# Returns data frame containing, for each simulated value, the left
# and right bounds of the confidence interval calculated by simulation.
left, right = biogeme.confidenceIntervals(b, 0.9)

wtpcar_left = (60 * left['WTP CAR time'] * left['weight']).mean()
wtpcar_right = (60 * right['WTP CAR time'] * right['weight']).mean()
print(
    f'Average WTP for car: {wtpcar:.3g} '
    f'CI:[{wtpcar_left:.3g}, {wtpcar_right:.3g}]'
)


# In this specific case, there are only two distinct values in the
# population: for workers and non workers
print(
    'Unique values: ',
    [f'{i:.3g}' for i in 60 * simulatedValues['WTP CAR time'].unique()],
)


def wtpForSubgroup(theFilter):
    """
    Check the value for groups of the population. Define a function that
    works for any filter to avoid repeating code.

    :param theFilter: pandas filter
    :type theFilter: numpy.Series(bool)

    :return: willingness-to-pay for car and confidence interval
    :rtype: tuple(float, float, float)
    """
    size = theFilter.sum()
    sim = simulatedValues[theFilter]
    totalWeight = sim['weight'].sum()
    weight = sim['weight'] * size / totalWeight
    _wtpcar = (60 * sim['WTP CAR time'] * weight).mean()
    _wtpcar_left = (60 * left[theFilter]['WTP CAR time'] * weight).mean()
    _wtpcar_right = (60 * right[theFilter]['WTP CAR time'] * weight).mean()
    return _wtpcar, _wtpcar_left, _wtpcar_right


# full time workers.
aFilter = database.data['OccupStat'] == 1
w, l, r = wtpForSubgroup(aFilter)
print(f'WTP car for workers: {w:.3g} CI:[{l:.3g}, {r:.3g}]')

# females
aFilter = database.data['Gender'] == 2
print(type(aFilter))
w, l, r = wtpForSubgroup(aFilter)
print(f'WTP car for females: {w:.3g} CI:[{l:.3g}, {r:.3g}]')

# males
aFilter = database.data['Gender'] == 1
w, l, r = wtpForSubgroup(aFilter)
print(f'WTP car for males: {w:.3g} CI:[{l:.3g}, {r:.3g}]')

# We plot the distribution of WTP in the population. In this case,
# there are only two values

plt.hist(
    60 * simulatedValues['WTP CAR time'], weights=simulatedValues['weight']
)
plt.xlabel('WTP (CHF/hour)')
plt.ylabel('Individuals')
plt.show()



"""File 04latentChoiceSeq.py

Choice model with the latent variable.
Mixture of logit.
Measurement equation for the indicators.
Sequential estimation.

:author: Michel Bierlaire, EPFL
:date: Tue Sep 10 08:13:18 2019

"""

import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.distributions as dist
import biogeme.results as res
import biogeme.messaging as msg
from biogeme.expressions import (
    Beta,
    DefineVariable,
    RandomVariable,
    exp,
    log,
    Integrate,
)

# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Read the estimates from the structural equation estimation
try:
    structResults = res.bioResults(pickleFile='02oneLatentOrdered.pickle')
except FileNotFoundError:
    print(
        'Run first the script 02oneLatentOrdered.py in order to generate the '
        'file 02oneLatentOrdered.pickle.'
    )
    sys.exit()
structBetas = structResults.getBetaValues()

### Variables

# Piecewise linear definition of income
ScaledIncome = DefineVariable(
    'ScaledIncome', CalculatedIncome / 1000, database
)
thresholds = [None, 4, 6, 8, 10, None]
piecewiseVariables = models.piecewiseVariables(ScaledIncome, thresholds)
formulaIncome = (
    structBetas['beta_ScaledIncome_lessthan_4'] * piecewiseVariables[0]
    + structBetas['beta_ScaledIncome_4_6'] * piecewiseVariables[1]
    + structBetas['beta_ScaledIncome_6_8'] * piecewiseVariables[2]
    + structBetas['beta_ScaledIncome_8_10'] * piecewiseVariables[3]
    + structBetas['beta_ScaledIncome_10_more'] * piecewiseVariables[4]
)


# Definition of other variables
age_65_more = DefineVariable('age_65_more', age >= 65, database)
moreThanOneCar = DefineVariable('moreThanOneCar', NbCar > 1, database)
moreThanOneBike = DefineVariable('moreThanOneBike', NbBicy > 1, database)
individualHouse = DefineVariable('individualHouse', HouseType == 1, database)
male = DefineVariable('male', Gender == 1, database)
haveChildren = DefineVariable(
    'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0, database
)
haveGA = DefineVariable('haveGA', GenAbST == 1, database)
highEducation = DefineVariable('highEducation', Education >= 6, database)

### Coefficients

coef_intercept = structBetas['coef_intercept']
coef_age_65_more = structBetas['coef_age_65_more']
coef_haveGA = structBetas['coef_haveGA']
coef_moreThanOneCar = structBetas['coef_moreThanOneCar']
coef_moreThanOneBike = structBetas['coef_moreThanOneBike']
coef_individualHouse = structBetas['coef_individualHouse']
coef_male = structBetas['coef_male']
coef_haveChildren = structBetas['coef_haveChildren']
coef_highEducation = structBetas['coef_highEducation']

### Latent variable: structural equation

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formulaIncome
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + sigma_s * omega
)

# Choice model
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', 0, None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', 0, None, None, 0)
BETA_DIST = Beta('BETA_DIST', 0, None, None, 0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', 0, None, 0, 0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', 0, None, 0, 0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', 0, None, None, 0)

# The coefficient of the latent variable should be initialized to
# something different from zero. If not, the algorithm may be trapped
# in a local optimum, and never change the value.
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', -0.01, None, None, 0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL', -0.01, None, None, 0)


TimePT_scaled = DefineVariable('TimePT_scaled', TimePT / 200, database)
TimeCar_scaled = DefineVariable('TimeCar_scaled', TimeCar / 200, database)
MarginalCostPT_scaled = DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10, database
)
CostCarCHF_scaled = DefineVariable(
    'CostCarCHF_scaled', CostCarCHF / 10, database
)
distance_km_scaled = DefineVariable(
    'distance_km_scaled', distance_km / 5, database
)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1, database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1, database)

### Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel)
condprob = models.logit(V, None, Choice)
# We integrate over omega using numerical integration
loglike = log(Integrate(condprob * density, 'omega'))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, loglike)
biogeme.modelName = '04latentChoiceSeq'

# Estimate the parameters
results = biogeme.estimate()

print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')
results.writeLaTeX()
print(f'LaTeX file: {results.data.latexFileName}')



"""File 05normalMixture_allAlgos.py

:author: Michel Bierlaire, EPFL
:date: Fri May  1 11:59:20 2020

 Example of a mixture of logit models, using Monte-Carlo integration.
 Three alternatives: Train, Car and Swissmetro
 SP data
"""

import pandas as pd
import biogeme.optimization as opt
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import Beta, DefineVariable, bioDraws, log, MonteCarlo

# Read the data
df = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to investigate the database. For example:
# print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
# remove = (((database.data.PURPOSE != 1) &
#           (database.data.PURPOSE != 3)) |
#          (database.data.CHOICE == 0))
# database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# Definition of new variables: adding columns to the database
CAR_AV_SP = DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0), database)
TRAIN_AV_SP = DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0), database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0, database)
TRAIN_COST_SCALED = DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100, database
)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0, database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100, database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100, database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100, database)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(prob))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, logprob, numberOfDraws=100000)

algos = {
    'scipy                   ': opt.scipy,
    'Line search             ': opt.newtonLineSearchForBiogeme,
    'Trust region (dogleg)   ': opt.newtonTrustRegionForBiogeme,
    'Trust region (cg)       ': opt.newtonTrustRegionForBiogeme,
    'LS-BFGS                 ': opt.bfgsLineSearchForBiogeme,
    'TR-BFGS                 ': opt.bfgsTrustRegionForBiogeme,
    'Simple bounds Newton fCG': opt.simpleBoundsNewtonAlgorithmForBiogeme,
    'Simple bounds BFGS fCG  ': opt.simpleBoundsNewtonAlgorithmForBiogeme,
    'Simple bounds hybrid fCG': opt.simpleBoundsNewtonAlgorithmForBiogeme,
    'Simple bounds Newton iCG': opt.simpleBoundsNewtonAlgorithmForBiogeme,
    'Simple bounds BFGS iCG  ': opt.simpleBoundsNewtonAlgorithmForBiogeme,
    'Simple bounds hybrid iCG': opt.simpleBoundsNewtonAlgorithmForBiogeme,
}

algoParameters = {
    'Trust region (dogleg)   ': {'dogleg': True},
    'Trust region (cg)       ': {'dogleg': False},
    'Simple bounds Newton fCG': {
        'proportionAnalyticalHessian': 1.0,
        'infeasibleConjugateGradient': False,
    },
    'Simple bounds BFGS fCG  ': {
        'proportionAnalyticalHessian': 0.0,
        'infeasibleConjugateGradient': False,
    },
    'Simple bounds hybrid fCG': {
        'proportionAnalyticalHessian': 0.5,
        'infeasibleConjugateGradient': False,
    },
    'Simple bounds Newton iCG': {
        'proportionAnalyticalHessian': 1.0,
        'infeasibleConjugateGradient': True,
    },
    'Simple bounds BFGS iCG  ': {
        'proportionAnalyticalHessian': 0.0,
        'infeasibleConjugateGradient': True,
    },
    'Simple bounds hybrid iCG': {
        'proportionAnalyticalHessian': 0.5,
        'infeasibleConjugateGradient': True,
    },
}

results = {}
msg = ''
for name, algo in algos.items():
    biogeme.modelName = f'05normalMixture_allAlgos_{name}'.strip()
    p = algoParameters.get(name)
    results[name] = biogeme.estimate(algorithm=algo, algoParameters=p)
    msg += (
        f'{name}\t{results[name].data.logLike:.2f}\t'
        f'{results[name].data.gradientNorm:.2g}\t'
        f'{results[name].data.optimizationMessages["Optimization time"]}'
        f'\t{results[name].data.optimizationMessages["Cause of termination"]}'
        f'\n'
    )

print('Algorithm\t\tloglike\t\tnormg\ttime\t\tdiagnostic')
print('+++++++++\t\t+++++++\t\t+++++\t++++\t\t++++++++++')
print(msg)

"""
Here are the results. Note that the draws are identical for all runs. Still, the algorithms
may converge to different solutions. Some algorithms obtain a solution with
B_TIME_S = 1.65 (LL = -5215.588),
and some obtain a solution with
B_TIME_S = -1.65 (LL = -5215.204).
Both are local optima of the likelihood function. As the draws are not exactly symmetric,
these solutions have different values for the objective functions. If the number of draws is
increased, the two local solutions will (asymptotically) become identical.

Algorithm               loglike         normg   time            diagnostic
+++++++++               +++++++         +++++   ++++            ++++++++++
scipy                           -5215.59        0.00024 0:00:26.443081  b'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
Line search                     -5215.59        3.8e-06 0:01:39.359296  Relative gradient = 9e-10 <= 6.1e-06
Trust region (dogleg)           -5215.20        0.0066  0:00:27.050893  Relative gradient = 1.5e-06 <= 6.1e-06
Trust region (cg)               -5215.20        0.013   0:00:33.720330  Relative gradient = 2.9e-06 <= 6.1e-06
LS-BFGS                         -5215.59        0.017   0:01:33.475099  Relative gradient = 2.9e-06 <= 6.1e-06
TR-BFGS                         -5215.59        0.018   0:01:22.466036  Relative gradient = 3.1e-06 <= 6.1e-06
Simple bounds Newton fCG        -5215.59        0.026   0:00:30.640365  Relative gradient = 5.1e-06 <= 6.1e-06
Simple bounds BFGS fCG          -5215.59        0.027   0:01:28.154424  Relative gradient = 4.7e-06 <= 6.1e-06
Simple bounds hybrid fCG        -5215.59        0.031   0:00:37.804337  Relative gradient = 3.9e-06 <= 6.1e-06
Simple bounds Newton iCG        -5215.20        0.017   0:00:26.515896  Relative gradient = 4e-06 <= 6.1e-06
Simple bounds BFGS iCG          -5215.59        0.027   0:01:28.327714  Relative gradient = 4.7e-06 <= 6.1e-06
Simple bounds hybrid iCG        -5215.20        0.0044  0:00:32.788981  Relative gradient = 8.9e-07 <= 6.1e-06
"""



"""File 15panelDiscreteBis.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 19:30:31 2019

 Example of a discrete mixture of logit models, also called latent class model.
 The datafile is organized as panel data.
 Here, we integrate before the discrete mixture to show that it is equivalent.
 Three alternatives: Train, Car and Swissmetro
 SP data
"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import (
    Beta,
    DefineVariable,
    bioDraws,
    PanelLikelihoodTrajectory,
    MonteCarlo,
    log,
)

# Read the data
df = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# They are organized as panel data. The variable ID identifies each individual.
database.panel("ID")

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
# remove = (((database.data.PURPOSE != 1) &
#           (database.data.PURPOSE != 3)) |
#          (database.data.CHOICE == 0))
# database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

# Parameters to be estimated. One version for each latent class.
numberOfClasses = 2
B_COST = [
    Beta(f'B_COST_class{i}', 0, None, None, 0) for i in range(numberOfClasses)
]

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME = [
    Beta(f'B_TIME_class{i}', 0, None, None, 0) for i in range(numberOfClasses)
]

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = [
    Beta(f'B_TIME_S_class{i}', 1, None, None, 0)
    for i in range(numberOfClasses)
]
B_TIME_RND = [
    B_TIME[i] + B_TIME_S[i] * bioDraws(f'B_TIME_RND_class{i}', 'NORMAL_ANTI')
    for i in range(numberOfClasses)
]

# We do the same for the constants, to address serial correlation.
ASC_CAR = [
    Beta(f'ASC_CAR_class{i}', 0, None, None, 0) for i in range(numberOfClasses)
]
ASC_CAR_S = [
    Beta(f'ASC_CAR_S_class{i}', 1, None, None, 0)
    for i in range(numberOfClasses)
]
ASC_CAR_RND = [
    ASC_CAR[i]
    + ASC_CAR_S[i] * bioDraws(f'ASC_CAR_RND_class{i}', 'NORMAL_ANTI')
    for i in range(numberOfClasses)
]

ASC_TRAIN = [
    Beta(f'ASC_TRAIN_class{i}', 0, None, None, 0)
    for i in range(numberOfClasses)
]
ASC_TRAIN_S = [
    Beta(f'ASC_TRAIN_S_class{i}', 1, None, None, 0)
    for i in range(numberOfClasses)
]
ASC_TRAIN_RND = [
    ASC_TRAIN[i]
    + ASC_TRAIN_S[i] * bioDraws(f'ASC_TRAIN_RND_class{i}', 'NORMAL_ANTI')
    for i in range(numberOfClasses)
]

ASC_SM = [
    Beta(f'ASC_SM_class{i}', 0, None, None, 1) for i in range(numberOfClasses)
]
ASC_SM_S = [
    Beta(f'ASC_SM_S_class{i}', 1, None, None, 0)
    for i in range(numberOfClasses)
]
ASC_SM_RND = [
    ASC_SM[i] + ASC_SM_S[i] * bioDraws(f'ASC_SM_RND_class{i}', 'NORMAL_ANTI')
    for i in range(numberOfClasses)
]

# Class memebership probability
PROB_class0 = Beta('PROB_class0', 0.5, 0, 1, 0)
PROB_class1 = 1 - PROB_class0

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# Definition of new variables: adding columns to the database
CAR_AV_SP = DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0), database)
TRAIN_AV_SP = DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0), database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0, database)
TRAIN_COST_SCALED = DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100, database
)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0, database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100, database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100, database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100, database)

# In class 0, it is assumed that the time coefficient is zero
B_TIME_RND[0] = 0

# Utility functions
V1 = [
    ASC_TRAIN_RND[i]
    + B_TIME_RND[i] * TRAIN_TT_SCALED
    + B_COST[i] * TRAIN_COST_SCALED
    for i in range(numberOfClasses)
]
V2 = [
    ASC_SM_RND[i] + B_TIME_RND[i] * SM_TT_SCALED + B_COST[i] * SM_COST_SCALED
    for i in range(numberOfClasses)
]
V3 = [
    ASC_CAR_RND[i] + B_TIME_RND[i] * CAR_TT_SCALED + B_COST[i] * CAR_CO_SCALED
    for i in range(numberOfClasses)
]
V = [{1: V1[i], 2: V2[i], 3: V3[i]} for i in range(numberOfClasses)]

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a discrete mixture of logit, with availability conditions
# We calculate the conditional probability for each class
prob = [
    MonteCarlo(PanelLikelihoodTrajectory(models.logit(V[i], av, CHOICE)))
    for i in range(numberOfClasses)
]

# Conditional to the random variables, likelihood for the individual.
probIndiv = PROB_class0 * prob[0] + PROB_class1 * prob[1]

# We integrate over the random variables using Monte-Carlo
logprob = log(probIndiv)

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, logprob, numberOfDraws=100000)
biogeme.modelName = '15panelDiscreteBis'

# Estimate the parameters.
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)



"""File 12panel_p.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 15 11:26:49 2020

 Example of a mixture of logit models, using Monte-Carlo integration.
 The datafile is organized as panel data.
 Three alternatives: Train, Car and Swissmetro
 SP data

The Swissmetro data is organized such that each row contains all the
responses of one individual.

"""
# pylint: disable=invalid-name, undefined-variable

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import (
    Beta,
    Variable,
    bioMultSum,
    DefineVariable,
    bioDraws,
    MonteCarlo,
    log,
    exp,
)

# Read the data
df = pd.read_csv('swissmetro_panel.dat', sep='\t')
database = db.Database('swissmetro', df)

# Number of observations for each individual. The are numbered from 0
# to 8 in the dat set.
nbrQuestions = 9

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Here we use the 'biogeme' way for backward compatibility
exclude = (PURPOSE != 1) * (PURPOSE != 3)
for q in range(nbrQuestions):
    exclude = exclude + (Variable(f'CHOICE_{q}') == 0)
database.remove(exclude > 0)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, 0, None, 0)

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of new variables
SM_COST = [Variable(f'SM_CO_{q}') * (GA == 0) for q in range(nbrQuestions)]

TRAIN_COST = [
    Variable(f'TRAIN_CO_{q}') * (GA == 0) for q in range(nbrQuestions)
]

# Definition of new variables: adding columns to the database
CAR_AV_SP = [
    DefineVariable(
        f'CAR_AV_SP_{q}', Variable(f'CAR_AV_{q}') * (SP != 0), database
    )
    for q in range(nbrQuestions)
]

TRAIN_AV_SP = [
    DefineVariable(
        f'TRAIN_AV_SP_{q}', Variable(f'TRAIN_AV_{q}') * (SP != 0), database
    )
    for q in range(nbrQuestions)
]

SM_AV = [Variable(f'SM_AV_{q}') for q in range(nbrQuestions)]

TRAIN_TT_SCALED = [
    DefineVariable(
        f'TRAIN_TT_SCALED_{q}', Variable(f'TRAIN_TT_{q}') / 100.0, database
    )
    for q in range(nbrQuestions)
]

TRAIN_COST_SCALED = [
    DefineVariable(f'TRAIN_COST_SCALED_{q}', TRAIN_COST[q] / 100, database)
    for q in range(nbrQuestions)
]

SM_TT_SCALED = [
    DefineVariable(
        f'SM_TT_SCALED_{q}', Variable(f'SM_TT_{q}') / 100.0, database
    )
    for q in range(nbrQuestions)
]

SM_COST_SCALED = [
    DefineVariable(f'SM_COST_SCALED_{q}', SM_COST[q] / 100, database)
    for q in range(nbrQuestions)
]

CAR_TT_SCALED = [
    DefineVariable(
        f'CAR_TT_SCALED_{q}', Variable(f'CAR_TT_{q}') / 100, database
    )
    for q in range(nbrQuestions)
]

CAR_CO_SCALED = [
    DefineVariable(
        f'CAR_CO_SCALED_{q}', Variable(f'CAR_CO_{q}') / 100, database
    )
    for q in range(nbrQuestions)
]

# Definition of the utility functions
V1 = [
    ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED[q] + B_COST * TRAIN_COST_SCALED[q]
    for q in range(nbrQuestions)
]
V2 = [
    ASC_SM + B_TIME_RND * SM_TT_SCALED[q] + B_COST * SM_COST_SCALED[q]
    for q in range(nbrQuestions)
]
V3 = [
    ASC_CAR + B_TIME_RND * CAR_TT_SCALED[q] + B_COST * CAR_CO_SCALED[q]
    for q in range(nbrQuestions)
]

# Associate utility functions with the numbering of alternatives
V = [{1: V1[q], 2: V2[q], 3: V3[q]} for q in range(nbrQuestions)]

av = [
    {1: TRAIN_AV_SP[q], 2: SM_AV[q], 3: CAR_AV_SP[q]}
    for q in range(nbrQuestions)
]


# Conditional to B_TIME_RND, the likelihood of one observation is
# given by the logit model (called the kernel)
obslogprob = [
    models.loglogit(V[q], av[q], Variable(f'CHOICE_{q}'))
    for q in range(nbrQuestions)
]

# Conditional to B_TIME_RND, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = exp(bioMultSum(obslogprob))

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
# logger.setGeneral()
logger.setDetailed()
# logger.setDebug()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, logprob, numberOfDraws=50000)
biogeme.modelName = '12panel_p'

# Estimate the parameters.
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)
